# Copyright (c) 2011, Dirk Thomas, Dorian Scholz, TU Darmstadt
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
#   * Redistributions of source code must retain the above copyright
#     notice, this list of conditions and the following disclaimer.
#   * Redistributions in binary form must reproduce the above
#     copyright notice, this list of conditions and the following
#     disclaimer in the documentation and/or other materials provided
#     with the distribution.
#   * Neither the name of the TU Darmstadt nor the names of its
#     contributors may be used to endorse or promote products derived
#     from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.


class PluginDescriptor(object):
    """Information about one Plugin."""

    def __init__(self, plugin_id, attributes=None):
        self._dict = {}
        self._dict['plugin_id'] = plugin_id
        self._dict['attributes'] = attributes or {}

    def plugin_id(self):
        return self._dict['plugin_id']

    def attributes(self):
        return self._dict['attributes']

    def action_attributes(self):
        return self._dict.get('action', {})

    def set_action_attributes(self, label, statustip=None, icon=None, icontype=None):
        self._dict['action'] = {
            'label': label,
            'statustip': statustip,
            'icon': icon,
            'icontype': icontype,
        }

    def groups(self):
        return self._dict.get('groups', [])

    def add_group_attributes(self, label, statustip=None, icon=None, icontype=None):
        if 'groups' not in self._dict:
            self._dict['groups'] = []
        self._dict['groups'].append({
            'label': label,
            'statustip': statustip,
            'icon': icon,
            'icontype': icontype,
        })



# Copyright (c) 2011, Dirk Thomas, Dorian Scholz, TU Darmstadt
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
#   * Redistributions of source code must retain the above copyright
#     notice, this list of conditions and the following disclaimer.
#   * Redistributions in binary form must reproduce the above
#     copyright notice, this list of conditions and the following
#     disclaimer in the documentation and/or other materials provided
#     with the distribution.
#   * Neither the name of the TU Darmstadt nor the names of its
#     contributors may be used to endorse or promote products derived
#     from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

from python_qt_binding.QtCore import QSignalMapper, Qt
from python_qt_binding.QtGui import QIcon
from python_qt_binding.QtWidgets import QAction, QToolBar, QWidget


class MinimizedDockWidgetsToolbar(QToolBar):

    max_label_length = 15

    def __init__(self, container_manager, parent=None):
        super(MinimizedDockWidgetsToolbar, self).__init__(parent=parent)
        self.setWindowTitle(self.tr('Minimized dock widgets'))
        self.setObjectName('MinimizedDockWidgetsToolbar')
        self.setToolButtonStyle(Qt.ToolButtonTextBesideIcon)
        self._container_manager = container_manager
        self._signal_mapper = QSignalMapper(self)
        self._signal_mapper.mapped[QWidget].connect(self._on_action_triggered)
        self._dock_widgets = {}

        self.hide()

    def addDockWidget(self, dock_widget):
        # remove action for same dock widget if exists
        self.removeDockWidget(dock_widget)

        icon = dock_widget.windowIcon()
        if icon.isNull():
            icon = QIcon.fromTheme('folder')
        title = dock_widget.windowTitle()
        action = QAction(icon, title, self)
        # truncate label if necessary
        if len(title) > MinimizedDockWidgetsToolbar.max_label_length:
            action.setToolTip(title)
            action.setIconText(title[0:MinimizedDockWidgetsToolbar.max_label_length] + '...')
        self._signal_mapper.setMapping(action, dock_widget)
        action.triggered.connect(self._signal_mapper.map)
        self._dock_widgets[dock_widget] = action
        self.addAction(action)

        self.show()

    def removeDockWidget(self, dock_widget):
        if dock_widget in self._dock_widgets:
            action = self._dock_widgets[dock_widget]
            self.removeAction(action)
            del self._dock_widgets[dock_widget]
            self._signal_mapper.removeMappings(action)

        if not self._dock_widgets:
            self.hide()

    def _on_action_triggered(self, dock_widget):
        # if the dock widget is nested inside a container also show the container
        # do this recursively for nested containers
        while True:
            parent = self._container_manager.get_container_of_dock_widget(dock_widget)
            if parent is None:
                break
            dock_widget.show()
            dock_widget = parent
        dock_widget.show()



# Copyright (c) 2011, Dirk Thomas, Dorian Scholz, TU Darmstadt
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
#   * Redistributions of source code must retain the above copyright
#     notice, this list of conditions and the following disclaimer.
#   * Redistributions in binary form must reproduce the above
#     copyright notice, this list of conditions and the following
#     disclaimer in the documentation and/or other materials provided
#     with the distribution.
#   * Neither the name of the TU Darmstadt nor the names of its
#     contributors may be used to endorse or promote products derived
#     from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

from python_qt_binding.QtCore import QMutex, QMutexLocker, QObject


class SettingsProxy(QObject):
    """Wrapper of a QSettings instance simplifying access of grouped data."""

    def __init__(self, qsettings):
        super(SettingsProxy, self).__init__()
        self.setObjectName('SettingsProxy')

        self._qsettings = qsettings
        self._mutex = QMutex(QMutex.Recursive)

    def all_keys(self, group):
        locker = QMutexLocker(self._mutex)  # noqa: F841
        self._qsettings.beginGroup(group)
        keys = self._qsettings.allKeys()
        self._qsettings.endGroup()
        return keys

#    def begin_read_array(self, group):

#    def begin_write_array(self, group):

    def child_groups(self, group):
        locker = QMutexLocker(self._mutex)  # noqa: F841
        self._qsettings.beginGroup(group)
        groups = self._qsettings.childGroups()
        self._qsettings.endGroup()
        return groups

    def child_keys(self, group):
        locker = QMutexLocker(self._mutex)  # noqa: F841
        self._qsettings.beginGroup(group)
        keys = self._qsettings.childKeys()
        self._qsettings.endGroup()
        return keys

    def contains(self, group, key):
        locker = QMutexLocker(self._mutex)  # noqa: F841
        self._qsettings.beginGroup(group)
        key_exists = self._qsettings.contains(key)
        self._qsettings.endGroup()
        return key_exists

#    def end_array(self):

    def remove(self, group, key):
        locker = QMutexLocker(self._mutex)  # noqa: F841
        self._qsettings.beginGroup(group)
        self._qsettings.remove(key)
        self._qsettings.endGroup()

#    def set_array_index(self, i):

    def set_value(self, group, key, value):
        locker = QMutexLocker(self._mutex)  # noqa: F841
        self._qsettings.beginGroup(group)
        self._qsettings.setValue(key, value)
        self._qsettings.endGroup()

    def value(self, group, key, default_value=None):
        locker = QMutexLocker(self._mutex)  # noqa: F841
        self._qsettings.beginGroup(group)
        v = self._qsettings.value(key, default_value)
        self._qsettings.endGroup()
        return v



# Copyright (c) 2011, Dirk Thomas, Dorian Scholz, TU Darmstadt
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
#   * Redistributions of source code must retain the above copyright
#     notice, this list of conditions and the following disclaimer.
#   * Redistributions in binary form must reproduce the above
#     copyright notice, this list of conditions and the following
#     disclaimer in the documentation and/or other materials provided
#     with the distribution.
#   * Neither the name of the TU Darmstadt nor the names of its
#     contributors may be used to endorse or promote products derived
#     from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import sys

from dbus.server import Server
from python_qt_binding import QT_BINDING
from python_qt_binding.QtCore import (QByteArray, QDataStream, qDebug, QIODevice,
                                      QProcess, QSignalMapper, Qt, qWarning)
from python_qt_binding.QtGui import QIcon, QToolBar, QX11EmbedContainer

from qt_gui.main import Main
from qt_gui.plugin_handler import PluginHandler
from qt_gui.plugin_handler_dbus_service import PluginHandlerDBusService
from qt_gui.settings_proxy_dbus_service import SettingsProxyDBusService


class PluginHandlerXEmbedContainer(PluginHandler):
    """
    Server part of the `PluginHandlerXEmbed`.

    Starts the plugin in a subprocess and provides the `PluginHandlerDBusService` through a
    peer-to-peer DBus connection.
    """

    _serial_number = 0

    def __init__(self, parent, main_window, instance_id,
                 application_context, container_manager, argv, dbus_object_path):
        super(PluginHandlerXEmbedContainer, self).__init__(
            parent, main_window, instance_id, application_context, container_manager, argv)
        self.setObjectName('PluginHandlerXEmbedContainer')

        self._dbus_object_path = dbus_object_path
        self._dbus_server = None
        self._dbus_container_service = None
        self._dbus_plugin_settings_service = None
        self._dbus_instance_settings_service = None

        self._process = None
        self._pid = None
        # mapping of widget object name to their embed container
        self._embed_containers = {}
        # mapping of toolbar object name to the toolbar
        self._embed_toolbars = {}
        self._signal_mapper_toolbars = QSignalMapper(self)
        self._signal_mapper_toolbars.mapped[str].connect(self._on_toolbar_orientation_changed)

    def _load(self):
        if not Main.main_filename:
            raise RuntimeError(
                'PluginHandlerXEmbedContainer._load() '
                'filename of initially started script is unknown')

        self._dbus_server = Server('tcp:bind=*')
        self._dbus_server.on_connection_added.append(self._add_dbus_connection)
        self._dbus_container_service = PluginHandlerDBusService(self, self._dbus_object_path)
        self._dbus_plugin_settings_service = SettingsProxyDBusService(
            self._dbus_object_path + '/plugin')
        self._dbus_instance_settings_service = SettingsProxyDBusService(
            self._dbus_object_path + '/instance')

        self._process = QProcess(self)
        self._process.setProcessChannelMode(QProcess.SeparateChannels)
        self._process.readyReadStandardOutput.connect(self._print_process_output)
        self._process.readyReadStandardError.connect(self._print_process_error)
        self._process.finished.connect(self._emit_close_plugin)
        # start python with unbuffered stdout/stderr so that the order of the output is retained
        cmd = sys.executable + ' -u'
        cmd += ' %s' % Main.main_filename
        cmd += ' --qt-binding=%s' % QT_BINDING
        cmd += ' --embed-plugin=%s --embed-plugin-serial=%s --embed-plugin-address=%s' % (
            self.instance_id().plugin_id, self.instance_id().serial_number,
            self._dbus_server.address)
        if self.argv():
            cmd += ' --args %s' % ' '.join(self.argv())
        # qDebug('PluginHandlerXEmbedContainer._load() starting command: %s' % cmd)
        self._process.start(cmd)
        started = self._process.waitForStarted(3000)
        if not started:
            self._dbus_container_service.remove_from_connection()
            self._dbus_plugin_settings_service.remove_from_connection()
            self._dbus_instance_settings_service.remove_from_connection()
            raise RuntimeError(
                'PluginHandlerXEmbedContainer._load() could not start '
                'subprocess in reasonable time')
        # QProcess.pid() has been added to PySide in 1.0.5
        if hasattr(self._process, 'pid'):
            self._pid = self._process.pid()
        else:
            # use serial number as a replacement for pid if not available
            self.__class__._serial_number = self._serial_number + 1
            self._pid = self._serial_number

        qDebug('PluginHandlerXEmbedContainer._load() started subprocess (#%s) for plugin "%s"' %
               (self._pid, str(self._instance_id)))
        # self._emit_load_completed is called asynchronous when client signals
        # finished loading via dbus

    def _add_dbus_connection(self, conn):
        self._dbus_container_service.add_to_connection(conn, self._dbus_object_path)
        self._dbus_plugin_settings_service.add_to_connection(
            conn, self._dbus_object_path + '/plugin')
        self._dbus_instance_settings_service.add_to_connection(
            conn, self._dbus_object_path + '/instance')

    def _print_process_output(self):
        self._print_process(self._process.readAllStandardOutput(), qDebug)

    def _print_process_error(self):
        self._print_process(self._process.readAllStandardError(), qWarning)

    def _print_process(self, data, method):
        # indent process output and prefix it with the pid
        lines = str(data).split('\n')
        if lines[-1] == '':
            lines.pop()
        for line in lines:
            method('    %d %s' % (self._pid, line))

    def load_completed(self, loaded, has_configuration):
        # TODO timer to detect no response
        exception = None if loaded else True
        self._plugin_has_configuration = has_configuration
        self._update_title_bars()
        self._emit_load_completed(exception)

    def _shutdown_plugin(self):
        qDebug('PluginHandlerXEmbedContainer._shutdown_plugin()')
        self._process.finished.disconnect(self._emit_close_plugin)
        self._dbus_container_service.shutdown_plugin()

    def emit_shutdown_plugin_completed(self):
        self._dbus_container_service.remove_from_connection()
        self._dbus_plugin_settings_service.remove_from_connection()
        self._dbus_instance_settings_service.remove_from_connection()

        self._process.close()
        self._process.waitForFinished(5000)
        if self._process.state() != QProcess.NotRunning:
            self._process.kill()
        self._process = None

        super(PluginHandlerXEmbedContainer, self).emit_shutdown_plugin_completed()

    def _unload(self):
        qDebug('PluginHandlerXEmbedContainer._unload()')
        self._emit_unload_completed()

    def _save_settings(self, plugin_settings, instance_settings):
        qDebug('PluginHandlerXEmbedContainer._save_settings()')
        self._dbus_plugin_settings_service.set_settings(plugin_settings)
        self._dbus_instance_settings_service.set_settings(instance_settings)
        self._dbus_container_service.save_settings()

    def emit_save_settings_completed(self):
        self._dbus_plugin_settings_service.set_settings(None)
        self._dbus_instance_settings_service.set_settings(None)
        super(PluginHandlerXEmbedContainer, self).emit_save_settings_completed()

    def _restore_settings(self, plugin_settings, instance_settings):
        qDebug('PluginHandlerXEmbedContainer._restore_settings()')
        self._dbus_plugin_settings_service.set_settings(plugin_settings)
        self._dbus_instance_settings_service.set_settings(instance_settings)
        self._dbus_container_service.restore_settings()

    def emit_restore_settings_completed(self):
        self._dbus_plugin_settings_service.set_settings(None)
        self._dbus_instance_settings_service.set_settings(None)
        super(PluginHandlerXEmbedContainer, self).emit_restore_settings_completed()

    def _trigger_configuration(self):
        self._dbus_container_service.trigger_configuration()

    def embed_widget(self, pid, widget_object_name):
        dock_widget = self._create_dock_widget()
        embed_container = QX11EmbedContainer(dock_widget)
        # embed_container.clientClosed.connect(self._emit_close_signal)
        self._add_dock_widget(dock_widget, embed_container)
        # update widget title is triggered by client after embedding
        self._embed_containers[widget_object_name] = embed_container
        return embed_container.winId()

    def update_embedded_widget_icon(self, widget_object_name, icon_str):
        embed_container = self._embed_containers[widget_object_name]
        # deserialize icon base64-encoded string
        ba = QByteArray.fromBase64(icon_str)
        s = QDataStream(ba, QIODevice.ReadOnly)
        icon = QIcon()
        s >> icon
        embed_container.setWindowIcon(icon)

    def update_embedded_widget_title(self, widget_object_name, title):
        embed_container = self._embed_containers[widget_object_name]
        embed_container.setWindowTitle(title)

    def unembed_widget(self, widget_object_name):
        embed_container = self._embed_containers[widget_object_name]
        self.remove_widget(embed_container)
        del self._embed_containers[widget_object_name]

    def embed_toolbar(self, pid, toolbar_object_name):
        toolbar = QToolBar()
        toolbar.setObjectName(toolbar_object_name)
        embed_container = QX11EmbedContainer(toolbar)
        toolbar.addWidget(embed_container)
        # embed_container.clientClosed.connect(self._emit_close_signal)
        self._add_toolbar(toolbar)
        self._embed_containers[toolbar_object_name] = embed_container
        # setup mapping to signal change of orientation to client
        self._embed_toolbars[toolbar_object_name] = toolbar
        self._signal_mapper_toolbars.setMapping(toolbar, toolbar_object_name)
        toolbar.orientationChanged.connect(self._signal_mapper_toolbars.map)
        return embed_container.winId()

    def _on_toolbar_orientation_changed(self, toolbar_object_name):
        embed_container = self._embed_containers[toolbar_object_name]
        toolbar = self._embed_toolbars[toolbar_object_name]
        self._dbus_container_service.toolbar_orientation_changed(
            embed_container.winId(), toolbar.orientation() == Qt.Horizontal)

    def unembed_toolbar(self, toolbar_object_name):
        embed_container = self._embed_containers[toolbar_object_name]
        del self._embed_containers[toolbar_object_name]
        del self._embed_toolbars[toolbar_object_name]
        self.remove_toolbar(embed_container)
        embed_container.close()



# Copyright (c) 2011, Dirk Thomas, Dorian Scholz, TU Darmstadt
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
#   * Redistributions of source code must retain the above copyright
#     notice, this list of conditions and the following disclaimer.
#   * Redistributions in binary form must reproduce the above
#     copyright notice, this list of conditions and the following
#     disclaimer in the documentation and/or other materials provided
#     with the distribution.
#   * Neither the name of the TU Darmstadt nor the names of its
#     contributors may be used to endorse or promote products derived
#     from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import webbrowser

from catkin_pkg.package import InvalidPackage, parse_package
from python_qt_binding.QtCore import QObject, Slot
from qt_gui.ros_package_helper import get_package_path


class HelpProvider(QObject):
    """Handler for the help action in the title bar of dock widgets."""

    def __init__(self):
        super(HelpProvider, self).__init__()

    @Slot(object)
    def plugin_help_request(self, plugin_descriptor):
        package_name = plugin_descriptor.attributes()['package_name']
        package_path = get_package_path(package_name)
        try:
            package = parse_package(package_path)
        except (InvalidPackage, IOError):
            return

        if len(package.urls) == 0:
            return
        url_str = package.urls[0].url
        for url in package.urls:
            if url.type == 'website':
                url_str = url.url
                break
        webbrowser.open(url_str)



# Copyright (c) 2011, Dirk Thomas, Dorian Scholz, TU Darmstadt
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
#   * Redistributions of source code must retain the above copyright
#     notice, this list of conditions and the following disclaimer.
#   * Redistributions in binary form must reproduce the above
#     copyright notice, this list of conditions and the following
#     disclaimer in the documentation and/or other materials provided
#     with the distribution.
#   * Neither the name of the TU Darmstadt nor the names of its
#     contributors may be used to endorse or promote products derived
#     from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import traceback

from python_qt_binding.QtCore import qCritical

from qt_gui.composite_plugin_provider import CompositePluginProvider


class RecursivePluginProvider(CompositePluginProvider):
    """
    Plugin provider which directly loads and returns all discovered plugins.

    Discovered plugins should be plugin providers themselves.
    """

    def __init__(self, plugin_provider):
        super(RecursivePluginProvider, self).__init__([])
        self.setObjectName('RecursivePluginProvider')

        self._plugin_provider = plugin_provider

    def discover(self, discovery_data):
        # discover plugins, which are providers themselves
        plugin_descriptors = self._plugin_provider.discover(discovery_data)

        # instantiate plugins
        plugin_providers = []
        for plugin_descriptor in plugin_descriptors:
            try:
                # pass None as PluginContext for PluginProviders
                instance = self._plugin_provider.load(plugin_descriptor.plugin_id(), None)
            except Exception:
                qCritical('RecursivePluginProvider.discover() loading plugin "%s" failed:\n%s' %
                          (str(plugin_descriptor.plugin_id()), traceback.format_exc()))
            else:
                if instance is not None:
                    plugin_providers.append(instance)

        # delegate discovery through instantiated plugin providers to base class
        self.set_plugin_providers(plugin_providers)
        return CompositePluginProvider.discover(self, discovery_data)



# Copyright (c) 2011, Dirk Thomas, Dorian Scholz, TU Darmstadt
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
#   * Redistributions of source code must retain the above copyright
#     notice, this list of conditions and the following disclaimer.
#   * Redistributions in binary form must reproduce the above
#     copyright notice, this list of conditions and the following
#     disclaimer in the documentation and/or other materials provided
#     with the distribution.
#   * Neither the name of the TU Darmstadt nor the names of its
#     contributors may be used to endorse or promote products derived
#     from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

from python_qt_binding.QtCore import QObject, Slot


class Settings(QObject):
    """Storage of key-value data with a QSettings-like interface."""

    def __init__(self, settings_proxy, group):
        super(Settings, self).__init__()
        self.setObjectName('Settings')

        self._settings_proxy = settings_proxy
        self._group = group

    def get_settings(self, group):
        prefix = self._group
        if prefix != '':
            prefix += '/'
        return Settings(self._settings_proxy, prefix + group)

    @Slot(result='QStringList')
    def all_keys(self):
        return self._settings_proxy.all_keys(self._group)

#    def begin_read_array(self):

#    def begin_write_array(self):

    @Slot(result='QStringList')
    def child_groups(self):
        return self._settings_proxy.child_groups(self._group)

    @Slot(result='QStringList')
    def child_keys(self):
        return self._settings_proxy.child_keys(self._group)

    @Slot(str, result=bool)
    def contains(self, key):
        return self._settings_proxy.contains(self._group, key)

#    def end_array(self):

    @Slot(str)
    def remove(self, key):
        self._settings_proxy.remove(self._group, key)

#    def set_array_index(self, i):

    @Slot(str, 'QVariant')
    def set_value(self, key, value):
        # work around for NoneType values via DBus
        if value is None:
            value = '__NoneType__'
        self._settings_proxy.set_value(self._group, key, value)

    @Slot(str, 'QVariant', result='QVariant')
    def value(self, key, default_value=None):
        # work around for passing NoneType (default_)values via DBus
        if default_value is None:
            default_value = '__NoneType__'
        value = self._settings_proxy.value(self._group, key, default_value)
        if value == '__NoneType__':
            value = None
        return value



# Copyright (c) 2011, Dirk Thomas, Dorian Scholz, TU Darmstadt
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
#   * Redistributions of source code must retain the above copyright
#     notice, this list of conditions and the following disclaimer.
#   * Redistributions in binary form must reproduce the above
#     copyright notice, this list of conditions and the following
#     disclaimer in the documentation and/or other materials provided
#     with the distribution.
#   * Neither the name of the TU Darmstadt nor the names of its
#     contributors may be used to endorse or promote products derived
#     from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

from python_qt_binding.QtCore import QObject
from python_qt_binding.QtWidgets import QAction, QMenu


class MenuManager(QObject):
    """Menu manager containing alphabetically ordered items and other items before and after."""

    def __init__(self, menu):
        super(MenuManager, self).__init__()
        self.setObjectName('MenuManager')

        self.menu = menu
        self._prefixes_separator = None
        self._ordered_items = []
        self._suffixes_separator = None

        # get already existing items from menu
        for action in menu.actions():
            menu = action.menu()
            if menu is not None:
                self._ordered_items.append(menu)
            else:
                self._ordered_items.append(action)

    def add_prefix(self, item):
        if self._prefixes_separator is None:
            before = self._ordered_items[0] if self._ordered_items else None
            if isinstance(before, QMenu):
                before = before.menuAction()
            self._prefixes_separator = self.menu.insertSeparator(before)
        self._insert_item(self._prefixes_separator, item)

    def add_item(self, new_item):
        for i, item in enumerate(self._ordered_items):
            if self._get_item_label(item) > self._get_item_label(new_item):
                self._insert_item(item, new_item)
                self._ordered_items.insert(i, new_item)
                return

        before = self._suffixes_separator or None
        self._insert_item(before, new_item)
        self._ordered_items.append(new_item)

    def add_suffix(self, item):
        if self._suffixes_separator is None:
            self._suffixes_separator = self.menu.addSeparator()
        self._insert_item(None, item)

    def count_items(self):
        return len(self._ordered_items)

    def contains_item(self, name):
        return self.get_item(name) is not None

    def get_item(self, name):
        if isinstance(name, QAction) or isinstance(name, QMenu):
            if name in self._ordered_items:
                return name
            return None
        for item in self._ordered_items:
            if self._get_item_label(item) == name:
                return item
        return None

    def contains_menu(self, name):
        return self.get_menu(name) is not None

    def get_menu(self, name):
        for item in self._ordered_items:
            if self._get_item_label(item) == name and isinstance(item, QMenu):
                return item
        return None

    def get_items(self):
        names = []
        for item in self._ordered_items:
            names.append(self._get_item_label(item))
        return names

    def set_item_checked(self, name, flag):
        item = self.get_item(name)
        if item is not None:
            item.setChecked(flag)

    def set_item_disabled(self, name, flag):
        item = self.get_item(name)
        if item is not None:
            item.setDisabled(flag)

    def remove_item(self, name):
        item = self.get_item(name)
        if item is not None:
            self.menu.removeAction(item)
            self._ordered_items.remove(item)
            item.deleteLater()

    def _insert_item(self, before, item):
        if isinstance(before, QMenu):
            before = before.menuAction()
        if isinstance(item, QAction):
            self.menu.insertAction(before, item)
        elif isinstance(item, QMenu):
            self.menu.insertMenu(before, item)
        elif item is None:
            self.menu.insertSeparator(before)
        else:
            raise UserWarning('unknown item type')

    def _get_item_label(self, item):
        if isinstance(item, QAction):
            return item.text()
        elif isinstance(item, QMenu):
            return item.title()
        else:
            raise UserWarning('unknown item type')



# Copyright (c) 2011, Dirk Thomas, Dorian Scholz, TU Darmstadt
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
#   * Redistributions of source code must retain the above copyright
#     notice, this list of conditions and the following disclaimer.
#   * Redistributions in binary form must reproduce the above
#     copyright notice, this list of conditions and the following
#     disclaimer in the documentation and/or other materials provided
#     with the distribution.
#   * Neither the name of the TU Darmstadt nor the names of its
#     contributors may be used to endorse or promote products derived
#     from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

from python_qt_binding.QtCore import qDebug, QEvent, QPoint, QRect, Qt
from python_qt_binding.QtGui import QMouseEvent
from python_qt_binding.QtWidgets import QApplication, QDockWidget

from qt_gui.dockable_main_window import DockableMainWindow
from qt_gui.reparent_event import ReparentEvent


class DockWidget(QDockWidget):
    """Widget with the capability to be reparented via drag-and-drop to any other main window."""

    def __init__(self, container_manager):
        super(DockWidget, self).__init__()
        self._container_manager = container_manager
        if self._container_manager is not None:
            self.event = self._event
        self._dragging_parent = None
        self._dragging_local_pos = None
        self._releasing_and_repressing_while_dragging = False
        self._main_windows = []

    def _event(self, e):
        if e.type() == QEvent.MouseButtonPress and e.button() == Qt.LeftButton:
            qDebug('%spress, rel=%s, global=%s, diff=%s' % (
                (' - pseudo ' if self._releasing_and_repressing_while_dragging else ''),
                e.pos(), e.globalPos(), e.globalPos() - self.pos()))

        if e.type() == QEvent.MouseButtonRelease and e.button() == Qt.LeftButton:
            qDebug('%srelease, rel=%s, global=%s, diff=%s' % (
                (' - pseudo ' if self._releasing_and_repressing_while_dragging else ''),
                e.pos(), e.globalPos(), e.globalPos() - self.pos()))

        # store local position when pressing button before starting the custom drag'n'drop
        # only allow when layout is not frozen
        if self._dragging_parent is None and \
                e.type() == QEvent.MouseButtonPress and \
                e.button() == Qt.LeftButton and \
                bool(self.features() & QDockWidget.DockWidgetMovable):
            self._dragging_local_pos = e.pos()

        if self._dragging_parent is None and \
                self._dragging_local_pos is not None and \
                e.type() == QEvent.Move and \
                QApplication.mouseButtons() & Qt.LeftButton:
            if self._widget_at(e.pos()) is not None:
                qDebug(
                    'DockWidget._event() start drag, dockwidget=%s, parent=%s, '
                    'floating=%s, pos=%s' % (
                        str(self), str(self.parent()),
                        str(self.isFloating()),
                        str(self._dragging_local_pos)))
                self._dragging_parent = self.parent()
                # ignore further mouse events so that the widget behind this dock widget
                # can be determined
                self.setAttribute(Qt.WA_TransparentForMouseEvents)

                # collect all main windows (except self.main_window) to re-implement
                # QApplication.widgetAt() in self._widget_at()
                self._main_windows = [self._container_manager.get_root_main_window()]
                for container in self._container_manager.get_containers():
                    if container == self:
                        continue
                    self._main_windows.append(container.main_window)

        # unset local position when releasing button even when custom drag'n'drop
        # has not been started
        if self._dragging_local_pos is not None and \
                e.type() == QEvent.MouseButtonRelease and \
                e.button() == Qt.LeftButton \
                and not self._releasing_and_repressing_while_dragging:
            self._dragging_local_pos = None

        if self._dragging_parent is not None and \
                e.type() == QEvent.MouseButtonRelease and \
                e.button() == Qt.LeftButton and \
                not self._releasing_and_repressing_while_dragging:
            qDebug('DockWidget._event() stop drag, dockwidget=%s, parent=%s\n' %
                   (self, self.parent()))
            self._dragging_parent = None
            self.setAttribute(Qt.WA_TransparentForMouseEvents, False)
            self._main_windows = []

        if self._dragging_parent is not None and \
                e.type() == QEvent.MouseMove and \
                e.buttons() & Qt.LeftButton and \
                not self._releasing_and_repressing_while_dragging:
            widget = self._widget_at(e.globalPos())
            new_parent = self._get_new_parent(widget)
            # print 'new_parent', new_parent, (new_parent.objectName() if new_parent else '')
            if new_parent is not None and new_parent != self.parent():
                self._releasing_and_repressing_while_dragging = True

                # schedule stop of pseudo drag'n'drop and let it complete
                mouse_release_event = QMouseEvent(
                    QEvent.MouseButtonRelease, self._dragging_local_pos,
                    e.globalPos(), Qt.LeftButton, Qt.NoButton, e.modifiers())
                QApplication.instance().postEvent(self, mouse_release_event)
                QApplication.sendPostedEvents()

                # schedule reparent to hovered main window and let it complete
                reparent_event = ReparentEvent(self, new_parent)
                QApplication.instance().postEvent(self._container_manager, reparent_event)
                QApplication.sendPostedEvents()

                # reenable mouse events to be able to receive upcoming pseudo mouse events
                self.setAttribute(Qt.WA_TransparentForMouseEvents, False)

                # schedule restart of pseudo drag'n'drop and let it complete
                mouse_repress_event = QMouseEvent(
                    QEvent.MouseButtonPress, self._dragging_local_pos, e.globalPos(),
                    Qt.LeftButton, Qt.LeftButton, e.modifiers())
                QApplication.instance().postEvent(self, mouse_repress_event)
                QApplication.sendPostedEvents()

                # schedule move to trigger dock widget drag'n'drop required for snapping and
                # showing rubber band and let it complete move forth...
                mouse_move_event = QMouseEvent(
                    QEvent.MouseMove,
                    self._dragging_local_pos,
                    e.globalPos() + QPoint(QApplication.startDragDistance(), 1),
                    Qt.NoButton,
                    Qt.LeftButton,
                    e.modifiers())
                QApplication.instance().postEvent(self, mouse_move_event)
                QApplication.sendPostedEvents()
                # ...and back
                mouse_move_event = QMouseEvent(
                    QEvent.MouseMove, self._dragging_local_pos, e.globalPos(),
                    Qt.NoButton, Qt.LeftButton, e.modifiers())
                QApplication.instance().postEvent(self, mouse_move_event)
                QApplication.sendPostedEvents()

                # restore attributes after repressing the button
                self.setAttribute(Qt.WA_TransparentForMouseEvents)

                self._releasing_and_repressing_while_dragging = False

        return super(DockWidget, self).event(e)

    def _get_new_parent(self, widget):
        from .dock_widget_container import DockWidgetContainer
        if isinstance(widget, DockWidgetContainer):
            if widget.isFloating():
                return None
            widget = widget.parent()
        while widget is not None:
            if isinstance(widget, DockableMainWindow):
                break
            if not callable(widget.parent):
                return None
            widget = widget.parent()
        return widget

    def _widget_at(self, global_point):
        # print '_widget_at()', global_point#, local_point
        widget = QApplication.widgetAt(global_point)
        # print '- widget', widget, (widget.objectName() if widget is not None else '')
        root_main_window = self._container_manager.get_root_main_window()

        # work around bug where root main window is detected when point is near but not inside it
        if widget == root_main_window and \
                not self._widget_contains(root_main_window, global_point):
            # print '- work around to large root main window'
            widget = None

        # work around bug where dock widget is floating and no widget is found
        if widget is None and self.isFloating():
            # determine all main windows which match the point
            overlapping = {}
            for main_window in self._main_windows:
                if self._widget_contains(main_window, global_point):
                    parent = main_window.parent()
                    is_floating = parent is None or parent.isFloating()
                    overlapping[main_window] = is_floating
            # print 'overlapping', overlapping

            if len(overlapping) == 1:
                # only found one candidate so pick it
                widget, _ = overlapping.popitem()
            elif len(overlapping) > 1:
                # try to determine which main window is the right one
                # determined docked main windows
                overlapping_docked = [mw for mw, floating in overlapping.items() if not floating]
                # print 'overlapping_docked', overlapping_docked

                # if at max one of the main windows is floating
                if len(overlapping_docked) >= len(overlapping) - 1:
                    # the floating main window is not considered because the docked ones are
                    # children of it

                    # consider all docked main windows and remove parent if both parent and child
                    # are in the list print 'considered docked main windows', overlapping_docked
                    parents = []
                    for mw1 in overlapping_docked:
                        # parent of the main window is a dock widget container
                        parent = mw1.parent()
                        if parent is None:
                            continue
                        # parent of the dock widget container is a main window
                        parent = parent.parent()
                        if parent is None:
                            continue
                        for mw2 in overlapping_docked:
                            if mw2 == parent:
                                parents.append(mw2)
                    for parent in parents:
                        overlapping_docked.remove(parent)
                    # print 'considered non-parent main windows', overlapping_docked

                    # if only one docked main window is found then pick it
                    if len(overlapping_docked) == 1:
                        # print '- pick single remaining main window'
                        widget = overlapping_docked[0]
                    else:
                        # print '- found multiple main windows - could not determine which one is
                        # on top'
                        pass
                # TODO any more heuristic possible?
                # if all remaining docked main windows have a most common ancestor use the
                # order of children() to determine topmost
        return widget

    def _widget_contains(self, widget, point):
        topleft = widget.mapToGlobal(widget.mapFromParent(widget.geometry().topLeft()))
        rect = QRect(topleft, widget.geometry().size())
        return rect.contains(point)

    def save_settings(self, settings):
        # print 'DockWidget.save_settings()', 'parent',
        # self._parent_container_serial_number(), 'settings group',
        # settings._group
        settings.set_value('parent', self._parent_container_serial_number())

        title_bar = self.titleBarWidget()
        title_bar.save_settings(settings)

    def restore_settings(self, settings):
        serial_number = settings.value('parent', None)
        # print 'DockWidget.restore_settings()', 'parent', serial_number,
        # 'settings group', settings._group
        if serial_number is not None:
            serial_number = int(serial_number)
        if self._parent_container_serial_number() != serial_number and \
                self._container_manager is not None:
            floating = self.isFloating()
            pos = self.pos()
            new_container = self._container_manager.get_container(serial_number)
            if new_container is not None:
                new_parent = new_container.main_window
            else:
                new_parent = self._container_manager.get_root_main_window()
            area = self.parent().dockWidgetArea(self)
            new_parent.addDockWidget(area, self)
            if floating:
                self.setFloating(floating)
                self.move(pos)

        title_bar = self.titleBarWidget()
        title_bar.restore_settings(settings)

    def _parent_container(self, dock_widget):
        from .dock_widget_container import DockWidgetContainer
        parent = dock_widget.parent()
        if parent is not None:
            parent = parent.parent()
            if isinstance(parent, DockWidgetContainer):
                return parent
        return None

    def _parent_container_serial_number(self):
        serial_number = None
        parent = self._parent_container(self)
        if parent is not None:
            serial_number = parent.serial_number()
        return serial_number



# Copyright (c) 2011, Dirk Thomas, Dorian Scholz, TU Darmstadt
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
#   * Redistributions of source code must retain the above copyright
#     notice, this list of conditions and the following disclaimer.
#   * Redistributions in binary form must reproduce the above
#     copyright notice, this list of conditions and the following
#     disclaimer in the documentation and/or other materials provided
#     with the distribution.
#   * Neither the name of the TU Darmstadt nor the names of its
#     contributors may be used to endorse or promote products derived
#     from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

from python_qt_binding.QtCore import qDebug, QObject, Qt

from qt_gui.dock_widget import DockWidget
from qt_gui.plugin_descriptor import PluginDescriptor
from qt_gui.reparent_event import ReparentEvent


class ContainerManager(QObject):
    """Manager of `DockWidgetContainer`s enabling reparenting to stored parent."""

    def __init__(self, root_main_window, parent=None):
        super(ContainerManager, self).__init__(parent)
        self._root_main_window = root_main_window
        self._container_descriptor = PluginDescriptor('__DockWidgetContainer')
        self._container_descriptor.set_action_attributes(
            self.tr('Container'), self.tr('Container for other dock widgets'),
            'folder-new', 'theme')

        self._containers = {}

    def get_root_main_window(self):
        return self._root_main_window

    def get_container_descriptor(self):
        return self._container_descriptor

    def add_to_plugin_menu(self, plugin_menu):
        plugin_menu.add_plugin_prefix(self._container_descriptor)

    def add_container(self, container):
        self._containers[container.serial_number()] = container

    def remove_container(self, container):
        del self._containers[container.serial_number()]

    def get_container(self, serial_number):
        if serial_number in self._containers:
            return self._containers[serial_number]
        return None

    def get_containers(self):
        return self._containers.values()

    def move_container_children_to_parent(self, container):
        floating = container.isFloating()
        for child in container.main_window.children():
            if isinstance(child, DockWidget):
                area = container.main_window.dockWidgetArea(child)
                container.parent().addDockWidget(area, child)
                if floating:
                    child.setFloating(floating)

    def get_container_of_dock_widget(self, dock_widget):
        for container in self._containers.values():
            for child in container.main_window.children():
                if child == dock_widget:
                    return container
        return None

    def restore_state_of_containers(self):
        for container in self._containers.values():
            container.restore_state()

    def event(self, e):
        if e.type() == ReparentEvent.reparent_event_type:
            qDebug('ContainerManager.event() reparent event: new parent=%s' %
                   e.new_parent.objectName())
            floating = e.dock_widget.isFloating()
            pos = e.dock_widget.pos()
            e.new_parent.addDockWidget(Qt.BottomDockWidgetArea, e.dock_widget)
            if floating:
                e.dock_widget.setFloating(floating)
                e.dock_widget.move(pos)
            e.accept()
            return True
        return super(ContainerManager, self).event(e)



# Copyright (c) 2011, Dirk Thomas, Dorian Scholz, TU Darmstadt
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
#   * Redistributions of source code must retain the above copyright
#     notice, this list of conditions and the following disclaimer.
#   * Redistributions in binary form must reproduce the above
#     copyright notice, this list of conditions and the following
#     disclaimer in the documentation and/or other materials provided
#     with the distribution.
#   * Neither the name of the TU Darmstadt nor the names of its
#     contributors may be used to endorse or promote products derived
#     from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import dbus
from dbus.service import Object

from python_qt_binding.QtCore import QObject, Qt, Signal


class PluginHandlerDBusService(Object):
    """DBus service for an indirect plugin handler, i.e. `PluginHandlerXEmbedContainer`."""

    class QueuedSignal(QObject):

        _signal = Signal()

        def __init__(self, callback):
            super(PluginHandlerDBusService.QueuedSignal, self).__init__()
            self._signal.connect(callback, Qt.QueuedConnection)

        def emit(self):
            self._signal.emit()

    def __init__(self, plugin_handler, object_path):
        super(PluginHandlerDBusService, self).__init__(object_path)
        self._plugin_handler = plugin_handler
        self._save_settings_completed = PluginHandlerDBusService.QueuedSignal(
            self._plugin_handler.emit_save_settings_completed)
        self._restore_settings_completed = PluginHandlerDBusService.QueuedSignal(
            self._plugin_handler.emit_restore_settings_completed)
        self._shutdown_plugin_completed = PluginHandlerDBusService.QueuedSignal(
            self._plugin_handler.emit_shutdown_plugin_completed)

    @dbus.service.method(
        'org.ros.qt_gui.PluginHandlerContainer', in_signature='bb', out_signature='')
    def load_completed(self, loaded, has_configuration):
        self._plugin_handler.load_completed(loaded, has_configuration)

    @dbus.service.method(
        'org.ros.qt_gui.PluginHandlerContainer', in_signature='is', out_signature='i')
    def embed_widget(self, pid, widget_object_name):
        return self._plugin_handler.embed_widget(pid, widget_object_name)

    @dbus.service.method(
        'org.ros.qt_gui.PluginHandlerContainer', in_signature='ss', out_signature='')
    def update_embedded_widget_icon(self, widget_object_name, icon):
        self._plugin_handler.update_embedded_widget_icon(widget_object_name, icon)

    @dbus.service.method(
        'org.ros.qt_gui.PluginHandlerContainer', in_signature='ss', out_signature='')
    def update_embedded_widget_title(self, widget_object_name, title):
        self._plugin_handler.update_embedded_widget_title(widget_object_name, title)

    @dbus.service.method(
        'org.ros.qt_gui.PluginHandlerContainer', in_signature='s', out_signature='')
    def unembed_widget(self, widget_object_name):
        self._plugin_handler.unembed_widget(widget_object_name)

    @dbus.service.method(
        'org.ros.qt_gui.PluginHandlerContainer', in_signature='is', out_signature='i')
    def embed_toolbar(self, pid, toolbar_object_name):
        return self._plugin_handler.embed_toolbar(pid, toolbar_object_name)

    @dbus.service.signal('org.ros.qt_gui.PluginHandlerContainer', signature='ib')
    def toolbar_orientation_changed(self, win_id, is_horizontal):
        # no implementation - any method call is relayed as a signal to the service client
        pass

    @dbus.service.method(
        'org.ros.qt_gui.PluginHandlerContainer', in_signature='s', out_signature='')
    def unembed_toolbar(self, toolbar_object_name):
        self._plugin_handler.unembed_toolbar(toolbar_object_name)

    @dbus.service.method('org.ros.qt_gui.PluginHandlerContainer', in_signature='',
                         out_signature='')
    def close_plugin(self):
        self._plugin_handler._emit_close_plugin()

    @dbus.service.method('org.ros.qt_gui.PluginHandlerContainer', in_signature='',
                         out_signature='')
    def reload_plugin(self):
        self._plugin_handler._emit_reload_signal()

    @dbus.service.signal('org.ros.qt_gui.PluginHandlerContainer', signature='')
    def shutdown_plugin(self):
        # no implementation - any method call is relayed as a signal to the service client
        pass

    @dbus.service.method('org.ros.qt_gui.PluginHandlerContainer', in_signature='',
                         out_signature='')
    def shutdown_plugin_completed(self):
        # schedule notification using queued signal, so that dbus does not wait
        # for the call to finish
        self._shutdown_plugin_completed.emit()

    @dbus.service.signal('org.ros.qt_gui.PluginHandlerContainer', signature='')
    def save_settings(self):
        # no implementation - any method call is relayed as a signal to the service client
        pass

    @dbus.service.method('org.ros.qt_gui.PluginHandlerContainer', signature='')
    def save_settings_completed(self):
        # schedule notification using queued signal, so that dbus does not wait
        # for the call to finish
        self._save_settings_completed.emit()

    @dbus.service.signal('org.ros.qt_gui.PluginHandlerContainer', signature='')
    def restore_settings(self):
        # no implementation - any method call is relayed as a signal to the service client
        pass

    @dbus.service.method('org.ros.qt_gui.PluginHandlerContainer', in_signature='',
                         out_signature='')
    def restore_settings_completed(self):
        # schedule notification using queued signal, so that dbus does not wait
        # for the call to finish
        self._restore_settings_completed.emit()

    @dbus.service.signal('org.ros.qt_gui.PluginHandlerContainer', signature='')
    def trigger_configuration(self):
        # no implementation - any method call is relayed as a signal to the service client
        pass



# Copyright (c) 2011, Dirk Thomas, Dorian Scholz, TU Darmstadt
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
#   * Redistributions of source code must retain the above copyright
#     notice, this list of conditions and the following disclaimer.
#   * Redistributions in binary form must reproduce the above
#     copyright notice, this list of conditions and the following
#     disclaimer in the documentation and/or other materials provided
#     with the distribution.
#   * Neither the name of the TU Darmstadt nor the names of its
#     contributors may be used to endorse or promote products derived
#     from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import dbus


class SettingsProxyDBusService(dbus.service.Object):
    """DBus service of a 'Settings' instance."""

    def __init__(self, object_path):
        super(SettingsProxyDBusService, self).__init__(object_path)
        self._settings_proxy = None
        self._group = None

    def set_settings(self, settings):
        if settings is not None:
            self._settings_proxy = settings._settings_proxy
            self._group = settings._group
        else:
            self._settings_proxy = None
            self._group = None

    @dbus.service.method('org.ros.qt_gui.Settings', in_signature='s', out_signature='av')
    def all_keys(self, group):
        return self._settings_proxy.all_keys(self._group + '/' + group)

    @dbus.service.method('org.ros.qt_gui.Settings', in_signature='s', out_signature='as')
    def child_groups(self, group):
        return self._settings_proxy.child_groups(self._group + '/' + group)

    @dbus.service.method('org.ros.qt_gui.Settings', in_signature='s', out_signature='as')
    def child_keys(self, group):
        return self._settings_proxy.child_keys(self._group + '/' + group)

    @dbus.service.method('org.ros.qt_gui.Settings', in_signature='ss', out_signature='b')
    def contains(self, group, key):
        return self._settings_proxy.contains(self._group + '/' + group, key)

    @dbus.service.method('org.ros.qt_gui.Settings', in_signature='ss', out_signature='')
    def remove(self, group, key):
        self._settings_proxy.remove(self._group + '/' + group, key)

    @dbus.service.method('org.ros.qt_gui.Settings', in_signature='ssv', out_signature='')
    def set_value(self, group, key, value):
        value = self._sanitize_value(value)
        self._settings_proxy.set_value(self._group + '/' + group, key, value)

    @dbus.service.method('org.ros.qt_gui.Settings', in_signature='ssv', out_signature='v')
    def value(self, group, key, default_value=None):
        return self._settings_proxy.value(self._group + '/' + group, key, default_value)

    def _sanitize_value(self, value):
        # transform DBus types to Python types to work with Pickle
        if isinstance(value, dbus.Boolean):
            value = bool(value)
        elif isinstance(value,
                        (dbus.Byte, dbus.Int16, dbus.Int32, dbus.Int64,
                            dbus.UInt16, dbus.UInt32, dbus.UInt64)):
            value = int(value)
        elif isinstance(value, dbus.Double):
            value = float(value)
        elif isinstance(value, dbus.String):
            value = str(value)
        elif isinstance(value, dbus.UTF8String):
            try:
                value = unicode(value)
            except NameError:
                value = str(value)
        return value



# Copyright (c) 2011, Dirk Thomas, Dorian Scholz, TU Darmstadt
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
#   * Redistributions of source code must retain the above copyright
#     notice, this list of conditions and the following disclaimer.
#   * Redistributions in binary form must reproduce the above
#     copyright notice, this list of conditions and the following
#     disclaimer in the documentation and/or other materials provided
#     with the distribution.
#   * Neither the name of the TU Darmstadt nor the names of its
#     contributors may be used to endorse or promote products derived
#     from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import traceback

from python_qt_binding.QtCore import qCritical, QEvent, QObject, qWarning, Slot

from qt_gui.plugin_context import PluginContext
from qt_gui.plugin_handler import PluginHandler


class PluginHandlerDirect(PluginHandler):
    """Handler for directly passing invocations between the framework and one `Plugin` instance."""

    def __init__(self, parent, main_window, instance_id,
                 application_context, container_manager, argv):
        super(PluginHandlerDirect, self).__init__(
            parent, main_window, instance_id, application_context, container_manager, argv)
        self.setObjectName('PluginHandlerDirect')
        self._context = None
        self._plugin = None

    def load(self, plugin_provider, callback=None):
        self._context = PluginContext(self)
        super(PluginHandlerDirect, self).load(plugin_provider, callback)

    def _load(self):
        self._plugin = self._plugin_provider.load(self._instance_id.plugin_id, self._context)
        if hasattr(self._plugin, 'has_configuration'):
            self._plugin_has_configuration = self._plugin.has_configuration()
        else:
            self._plugin_has_configuration = hasattr(self._plugin, 'trigger_configuration')
        self._update_title_bars()
        self._emit_load_completed()

    def _emit_load_completed(self, exception=None):
        if exception is None and hasattr(self._plugin, 'installEventFilter'):
            # emit close_signal when deferred delete event for plugin is received
            self._plugin.installEventFilter(self)
        super(PluginHandlerDirect, self)._emit_load_completed(exception)

    def eventFilter(self, watched, event):
        if event.type() == QEvent.DeferredDelete:
            # TOOD: check if ignore() is necessary
            event.ignore()
            self.close_signal.emit(str(self._instance_id))
            return True
        return QObject.eventFilter(self, watched, event)

    def shutdown_plugin(self, callback):
        if hasattr(self._plugin, 'removeEventFilter'):
            self._plugin.removeEventFilter(self)
        super(PluginHandlerDirect, self).shutdown_plugin(callback)
        if hasattr(self._plugin, 'deleteLater'):
            self._plugin.deleteLater()

    def _shutdown_plugin(self):
        if hasattr(self._plugin, 'shutdown_plugin'):
            try:
                self._plugin.shutdown_plugin()
            except Exception:
                qCritical(
                    'PluginHandlerDirect._shutdown_plugin() plugin "%s" raised an '
                    'exception:\n%s' % (str(self._instance_id), traceback.format_exc()))
        self.emit_shutdown_plugin_completed()

    def _delete_widget(self, widget):
        # only delete widgets which are not at the same time the plugin
        if widget != self._plugin:
            super(PluginHandlerDirect, self)._delete_widget(widget)

    def _unload(self):
        self._plugin_provider.unload(self._plugin)
        self._plugin = None
        self._emit_unload_completed()

    def _save_settings(self, plugin_settings, instance_settings):
        if hasattr(self._plugin, 'save_settings'):
            plugin_settings_plugin = plugin_settings.get_settings('plugin')
            instance_settings_plugin = instance_settings.get_settings('plugin')
            try:
                self._plugin.save_settings(plugin_settings_plugin, instance_settings_plugin)
            except Exception:
                qCritical(
                    'PluginHandlerDirect._save_settings() plugin "%s" raised an exception:\n%s' %
                    (str(self._instance_id), traceback.format_exc()))
        self.emit_save_settings_completed()

    def _restore_settings(self, plugin_settings, instance_settings):
        if hasattr(self._plugin, 'restore_settings'):
            plugin_settings_plugin = plugin_settings.get_settings('plugin')
            instance_settings_plugin = instance_settings.get_settings('plugin')
            try:
                self._plugin.restore_settings(plugin_settings_plugin, instance_settings_plugin)
            except Exception:
                qCritical(
                    'PluginHandlerDirect._restore_settings() plugin "%s" raised an '
                    'exception:\n%s' % (str(self._instance_id), traceback.format_exc()))
        self.emit_restore_settings_completed()

    # pointer to QWidget must be used for PySide to work (at least with 1.0.1)
    @Slot('QWidget*')
    def add_widget(self, widget):
        if widget in self._widgets:
            qWarning('PluginHandlerDirect.add_widget() widget "%s" already added' %
                     widget.objectName())
            return
        dock_widget = self._create_dock_widget()
        self._add_dock_widget(dock_widget, widget)

    # pointer to QToolBar must be used for PySide to work (at least with 1.0.1)
    @Slot('QToolBar*')
    def add_toolbar(self, toolbar):
        if toolbar in self._toolbars:
            qWarning('PluginHandlerDirect.add_toolbar() toolbar "%s" already added' %
                     toolbar.objectName())
            return
        self._add_toolbar(toolbar)

    @Slot()
    def close_plugin(self):
        # only non-standalone plugins are closable
        if self._application_context.options.standalone_plugin is None:
            self._emit_close_plugin()

    @Slot()
    def reload_plugin(self):
        self._emit_reload_signal()



# Copyright (c) 2011, Dirk Thomas, Dorian Scholz, TU Darmstadt
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
#   * Redistributions of source code must retain the above copyright
#     notice, this list of conditions and the following disclaimer.
#   * Redistributions in binary form must reproduce the above
#     copyright notice, this list of conditions and the following
#     disclaimer in the documentation and/or other materials provided
#     with the distribution.
#   * Neither the name of the TU Darmstadt nor the names of its
#     contributors may be used to endorse or promote products derived
#     from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import os
import traceback

from dbus import Interface
from dbus.connection import Connection
from python_qt_binding.QtCore import (QByteArray, qCritical, QDataStream,
                                      qDebug, QIODevice, Qt, qWarning, Slot)
from python_qt_binding.QtGui import QVBoxLayout, QX11EmbedWidget

from qt_gui.plugin_handler_direct import PluginHandlerDirect
from qt_gui.settings import Settings
from qt_gui.window_changed_signaler import WindowChangedSignaler


class PluginHandlerXEmbedClient(PluginHandlerDirect):
    """
    Client part of the `PluginHandlerXEmbed`.

    Client utilizes the `PluginHandlerDBusService` of the `PluginHandlerXEmbedContainer` through a
    peer-to-peer DBus connection.
    """

    def __init__(self, parent, main_window, instance_id,
                 application_context, container_manager, argv, dbus_object_path):
        super(PluginHandlerXEmbedClient, self).__init__(
            parent, main_window, instance_id, application_context, container_manager, argv)
        self.setObjectName('PluginHandlerXEmbedClient')
        self._dbus_object_path = dbus_object_path
        self._remote_container = None
        self._remote_plugin_settings = None
        self._remote_instance_settings = None
        # mapping of added widgets to their embed widget and WindowChangedSignaler
        self._embed_widgets = {}

    def _load(self):
        conn = Connection(self._application_context.options.embed_plugin_address)
        proxy = conn.get_object(None, self._dbus_object_path)
        self._remote_container = Interface(proxy, 'org.ros.qt_gui.PluginHandlerContainer')
        self._remote_container.connect_to_signal('shutdown_plugin', self._shutdown_plugin)
        self._remote_container.connect_to_signal('save_settings', self._save_settings_from_remote)
        self._remote_container.connect_to_signal(
            'restore_settings', self._restore_settings_from_remote)
        self._remote_container.connect_to_signal(
            'trigger_configuration', self._trigger_configuration)
        self._remote_container.connect_to_signal(
            'toolbar_orientation_changed', self._toolbar_orientation_changed)

        proxy = conn.get_object(None, self._dbus_object_path + '/plugin')
        self._remote_plugin_settings = Interface(proxy, 'org.ros.qt_gui.Settings')
        proxy = conn.get_object(None, self._dbus_object_path + '/instance')
        self._remote_instance_settings = Interface(proxy, 'org.ros.qt_gui.Settings')

        super(PluginHandlerXEmbedClient, self)._load()

    def _emit_load_completed(self, exception=None):
        # signal failed loading before emitting signal, as it might not be possible afterwards
        if exception is not None:
            self._remote_container.load_completed(False, False)
        super(PluginHandlerXEmbedClient, self)._emit_load_completed(exception)
        # signal successful loading after emitting signal, for better message order
        if exception is None:
            self._remote_container.load_completed(True, self._plugin_has_configuration)

    def shutdown_plugin(self, callback):
        # this method should never be called for embedded clients
        assert False

    def emit_shutdown_plugin_completed(self):
        self._remote_container.shutdown_plugin_completed()

    def save_settings(self, plugin_settings, instance_settings, callback=None):
        # this method should never be called for embedded clients
        assert False

    def _save_settings_from_remote(self):
        qDebug('PluginHandlerXEmbedClient._save_settings_from_remote()')
        try:
            plugin_settings = Settings(self._remote_plugin_settings, '')
            instance_settings = Settings(self._remote_instance_settings, '')
            self._save_settings(plugin_settings, instance_settings)
        except Exception:
            qCritical(
                'PluginHandlerXEmbedClient._save_settings_from_remote() plugin "%s" '
                'raised an exception:\n%s' % (str(self._instance_id), traceback.format_exc()))
            self.emit_save_settings_completed()

    def emit_save_settings_completed(self):
        self._remote_container.save_settings_completed()

    def restore_settings(self, plugin_settings, instance_settings, callback=None):
        # this method should never be called for embedded clients
        assert False

    def _restore_settings_from_remote(self):
        qDebug('PluginHandlerXEmbedClient._restore_settings_from_remote()')
        try:
            plugin_settings = Settings(self._remote_plugin_settings, '')
            instance_settings = Settings(self._remote_instance_settings, '')
            self._restore_settings(plugin_settings, instance_settings)
        except Exception:
            qCritical(
                'PluginHandlerXEmbedClient._restore_settings_from_remote() '
                'plugin "%s" raised an exception:\n%s' %
                (str(self._instance_id), traceback.format_exc()))
            self.emit_restore_settings_completed()

    def emit_restore_settings_completed(self):
        self._remote_container.restore_settings_completed()

    # pointer to QWidget must be used for PySide to work (at least with 1.0.1)
    @Slot('QWidget*')
    def add_widget(self, widget):
        if widget in self._embed_widgets:
            qWarning('PluginHandlerXEmbedClient.add_widget() widget "%s" already added' %
                     widget.objectName())
            return
        embed_widget = QX11EmbedWidget()
        layout = QVBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        layout.addWidget(widget)
        embed_widget.setLayout(layout)

        # close embed widget when container is closed
        # TODO necessary?
        # embed_widget.containerClosed.connect(embed_widget.close)

        embed_container_window_id = self._remote_container.embed_widget(
            os.getpid(), widget.objectName())
        embed_widget.embedInto(embed_container_window_id)

        signaler = WindowChangedSignaler(widget, widget)
        signaler.window_icon_changed_signal.connect(self._on_embed_widget_icon_changed)
        signaler.window_title_changed_signal.connect(self._on_embed_widget_title_changed)
        self._embed_widgets[widget] = embed_widget, signaler
        # trigger to update initial window icon and title
        signaler.window_icon_changed_signal.emit(widget)
        signaler.window_title_changed_signal.emit(widget)

        embed_widget.show()

    def _on_embed_widget_icon_changed(self, widget):
        # serialize icon base64-encoded string
        ba = QByteArray()
        s = QDataStream(ba, QIODevice.WriteOnly)
        s << widget.windowIcon()
        icon_str = str(ba.toBase64())
        self._remote_container.update_embedded_widget_icon(widget.objectName(), icon_str)

    def _on_embed_widget_title_changed(self, widget):
        self._remote_container.update_embedded_widget_title(
            widget.objectName(), widget.windowTitle())

    # pointer to QWidget must be used for PySide to work (at least with 1.0.1)
    @Slot('QWidget*')
    def remove_widget(self, widget):
        embed_widget, signaler = self._embed_widgets[widget]
        del self._embed_widgets[widget]
        signaler.window_icon_changed_signal.disconnect(self._on_embed_widget_icon_changed)
        signaler.window_title_changed_signal.disconnect(self._on_embed_widget_title_changed)
        self._remote_container.unembed_widget(widget.objectName())
        # do not delete the widget, only the embed widget
        widget.setParent(None)
        embed_widget.deleteLater()
        # triggering close after last widget and toolbar is closed is handled by the container

    # pointer to QToolBar must be used for PySide to work (at least with 1.0.1)
    @Slot('QToolBar*')
    def add_toolbar(self, toolbar):
        if toolbar in self._embed_widgets:
            qWarning('PluginHandlerXEmbedClient.add_toolbar() toolbar "%s" already added' %
                     toolbar.objectName())
            return
        embed_widget = QX11EmbedWidget()
        layout = QVBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        layout.addWidget(toolbar)
        embed_widget.setLayout(layout)

        # close embed widget when container is closed
        # TODO necessary?
        # embed_widget.containerClosed.connect(embed_widget.close)
        def foo():
            print('embed_widget.containerClosed')
        embed_widget.containerClosed.connect(foo)

        embed_container_window_id = self._remote_container.embed_toolbar(
            os.getpid(), toolbar.objectName())
        embed_widget.embedInto(embed_container_window_id)

        self._embed_widgets[toolbar] = embed_widget, None

        embed_widget.show()

    def _toolbar_orientation_changed(self, win_id, is_horizontal):
        for toolbar, (embed_widget, _) in self._embed_widgets.items():
            if embed_widget.containerWinId() == win_id:
                toolbar.setOrientation(Qt.Horizontal if is_horizontal else Qt.Vertical)
                break

    # pointer to QToolBar must be used for PySide to work (at least with 1.0.1)
    @Slot('QToolBar*')
    def remove_toolbar(self, toolbar):
        embed_widget, _ = self._embed_widgets[toolbar]
        del self._embed_widgets[toolbar]
        self._remote_container.unembed_widget(toolbar.objectName())
        # do not delete the toolbar, only the embed widget
        toolbar.setParent(None)
        embed_widget.deleteLater()
        # triggering close after last widget and toolbar is closed is handled by the container

    def _emit_close_plugin(self):
        self._remote_container.close_plugin()



# Copyright (c) 2011, Dirk Thomas, Dorian Scholz, TU Darmstadt
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
#   * Redistributions of source code must retain the above copyright
#     notice, this list of conditions and the following disclaimer.
#   * Redistributions in binary form must reproduce the above
#     copyright notice, this list of conditions and the following
#     disclaimer in the documentation and/or other materials provided
#     with the distribution.
#   * Neither the name of the TU Darmstadt nor the names of its
#     contributors may be used to endorse or promote products derived
#     from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

from python_qt_binding.QtCore import QObject


class Plugin(QObject):
    """
    Interface for Python plugins.

    User-defined plugins may either subclass `qt_gui.plugin.Plugin` or according to duck typing
    implement only the needed methods.
    """

    def __init__(self, context):
        """Instantiate the plugin and pass the `PluginContext`."""
        super(Plugin, self).__init__(context)
        self.setObjectName('Plugin')

    def shutdown_plugin(self):
        """Shutdown and clean up the plugin before unloading."""
        pass

    def save_settings(self, plugin_settings, instance_settings):
        """
        Save the intrinsic state of the plugin to the plugin or instance specific `Settings`.

        @param plugin_settings: The plugin-specific settings
        @type plugin_settings: qt_gui.settings.Settings
        @param instance_settings: The instance-specific settings
        @type instance_settings: qt_gui.settings.Settings
        """
        pass

    def restore_settings(self, plugin_settings, instance_settings):
        """
        Restore the intrinsic state of the plugin from the plugin or instance specific `Settings`.

        @param plugin_settings: The plugin-specific settings
        @type plugin_settings: qt_gui.settings.Settings
        @param instance_settings: The instance-specific settings
        @type instance_settings: qt_gui.settings.Settings
        """
        pass

    # def trigger_configuration(self):
        # """
        # Trigger a configuration request from the title bar of one of the dock widgets.
        # If this method is available the `DockWidgetTitleBar` will show the configuration action.
        # """
        # pass



# Copyright (c) 2011, Dirk Thomas, Dorian Scholz, TU Darmstadt
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
#   * Redistributions of source code must retain the above copyright
#     notice, this list of conditions and the following disclaimer.
#   * Redistributions in binary form must reproduce the above
#     copyright notice, this list of conditions and the following
#     disclaimer in the documentation and/or other materials provided
#     with the distribution.
#   * Neither the name of the TU Darmstadt nor the names of its
#     contributors may be used to endorse or promote products derived
#     from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import json
import os

from python_qt_binding import loadUi
from python_qt_binding.QtCore import QByteArray, qDebug, QObject, QSignalMapper, Signal, Slot
from python_qt_binding.QtGui import QIcon, QValidator
from python_qt_binding.QtWidgets import QAction, QFileDialog, QInputDialog, QMessageBox

from qt_gui.menu_manager import MenuManager
from qt_gui.settings import Settings
from qt_gui.settings_proxy import SettingsProxy


def is_string(s):
    """Check if the argument is a string which works for both Python 2 and 3."""
    try:
        return isinstance(s, basestring)
    except NameError:
        return isinstance(s, str)


class PerspectiveManager(QObject):
    """Manager for perspectives associated with specific sets of `Settings`."""

    perspective_changed_signal = Signal(str)
    save_settings_signal = Signal(Settings, Settings)
    restore_settings_signal = Signal(Settings, Settings)
    restore_settings_without_plugin_changes_signal = Signal(Settings, Settings)

    HIDDEN_PREFIX = '@'

    def __init__(self, settings, application_context):
        super(PerspectiveManager, self).__init__()
        self.setObjectName('PerspectiveManager')

        self._qtgui_path = application_context.qtgui_path

        self._settings_proxy = SettingsProxy(settings)
        self._global_settings = Settings(self._settings_proxy, 'global')
        self._perspective_settings = None
        self._create_perspective_dialog = None

        self._menu_manager = None
        self._perspective_mapper = None

        # get perspective list from settings
        self.perspectives = self._settings_proxy.value('', 'perspectives', [])
        if is_string(self.perspectives):
            self.perspectives = [self.perspectives]

        self._current_perspective = None
        self._remove_action = None

        self._callback = None
        self._callback_args = []

        self._file_path = os.getcwd()

        if application_context.provide_app_dbus_interfaces:
            from qt_gui.perspective_manager_dbus_interface import PerspectiveManagerDBusInterface
            self._dbus_server = PerspectiveManagerDBusInterface(self, application_context)

    def set_menu(self, menu):
        self._menu_manager = MenuManager(menu)
        self._perspective_mapper = QSignalMapper(menu)
        self._perspective_mapper.mapped[str].connect(self.switch_perspective)

        # generate menu
        create_action = QAction('&Create perspective...', self._menu_manager.menu)
        create_action.setIcon(QIcon.fromTheme('list-add'))
        create_action.triggered.connect(self._on_create_perspective)
        self._menu_manager.add_suffix(create_action)

        self._remove_action = QAction('&Remove perspective...', self._menu_manager.menu)
        self._remove_action.setEnabled(False)
        self._remove_action.setIcon(QIcon.fromTheme('list-remove'))
        self._remove_action.triggered.connect(self._on_remove_perspective)
        self._menu_manager.add_suffix(self._remove_action)

        self._menu_manager.add_suffix(None)

        import_action = QAction('&Import...', self._menu_manager.menu)
        import_action.setIcon(QIcon.fromTheme('document-open'))
        import_action.triggered.connect(self._on_import_perspective)
        self._menu_manager.add_suffix(import_action)

        export_action = QAction('&Export...', self._menu_manager.menu)
        export_action.setIcon(QIcon.fromTheme('document-save-as'))
        export_action.triggered.connect(self._on_export_perspective)
        self._menu_manager.add_suffix(export_action)

        # add perspectives to menu
        for name in self.perspectives:
            if not name.startswith(self.HIDDEN_PREFIX):
                self._add_perspective_action(name)

    def set_perspective(self, name, hide_and_without_plugin_changes=False):
        if name is None:
            name = self._settings_proxy.value('', 'current-perspective', 'Default')
        elif hide_and_without_plugin_changes:
            name = self.HIDDEN_PREFIX + name
        self.switch_perspective(name, save_before=not hide_and_without_plugin_changes,
                                without_plugin_changes=hide_and_without_plugin_changes)

    @Slot(str)
    @Slot(str, bool)
    @Slot(str, bool, bool)
    def switch_perspective(
            self, name, settings_changed=True, save_before=True, without_plugin_changes=False):
        if save_before and \
                self._global_settings is not None and \
                self._perspective_settings is not None:
            self._callback = self._switch_perspective
            self._callback_args = [name, settings_changed, save_before]
            self.save_settings_signal.emit(self._global_settings, self._perspective_settings)
        else:
            self._switch_perspective(name, settings_changed, save_before, without_plugin_changes)

    def _switch_perspective(
            self, name, settings_changed, save_before, without_plugin_changes=False):
        # convert from unicode
        name = str(name.replace('/', '__'))

        qDebug('PerspectiveManager.switch_perspective() switching to perspective "%s"' % name)
        if self._current_perspective is not None and self._menu_manager is not None:
            self._menu_manager.set_item_checked(self._current_perspective, False)
            self._menu_manager.set_item_disabled(self._current_perspective, False)

        # create perspective if necessary
        if name not in self.perspectives:
            self._create_perspective(name, clone_perspective=False)

        # update current perspective
        self._current_perspective = name
        if self._menu_manager is not None:
            self._menu_manager.set_item_checked(self._current_perspective, True)
            self._menu_manager.set_item_disabled(self._current_perspective, True)
        if not self._current_perspective.startswith(self.HIDDEN_PREFIX):
            self._settings_proxy.set_value('', 'current-perspective', self._current_perspective)
        self._perspective_settings = self._get_perspective_settings(self._current_perspective)

        # emit signals
        self.perspective_changed_signal.emit(self._current_perspective.lstrip(self.HIDDEN_PREFIX))
        if settings_changed:
            if not without_plugin_changes:
                self.restore_settings_signal.emit(
                    self._global_settings, self._perspective_settings)
            else:
                self.restore_settings_without_plugin_changes_signal.emit(
                    self._global_settings, self._perspective_settings)

    def save_settings_completed(self):
        if self._callback is not None:
            callback = self._callback
            callback_args = self._callback_args
            self._callback = None
            self._callback_args = []
            callback(*callback_args)

    def _get_perspective_settings(self, perspective_name):
        return Settings(self._settings_proxy, 'perspective/%s' % perspective_name)

    def _on_create_perspective(self):
        name = self._choose_new_perspective_name()
        if name is not None:
            clone_perspective = self._create_perspective_dialog.clone_checkbox.isChecked()
            self._create_perspective(name, clone_perspective)
            self.switch_perspective(
                name, settings_changed=not clone_perspective, save_before=False)

    def _choose_new_perspective_name(self, show_cloning=True):
        # input dialog for new perspective name
        if self._create_perspective_dialog is None:
            ui_file = os.path.join(self._qtgui_path,
                                   'share', 'qt_gui', 'resource', 'perspective_create.ui')
            self._create_perspective_dialog = loadUi(ui_file)

            # custom validator preventing forward slashs
            class CustomValidator(QValidator):

                def __init__(self, parent=None):
                    super(CustomValidator, self).__init__(parent)

                def fixup(self, value):
                    value = value.replace('/', '')

                def validate(self, value, pos):
                    if value.find('/') != -1:
                        pos = value.find('/')
                        return (QValidator.Invalid, value, pos)
                    if value == '':
                        return (QValidator.Intermediate, value, pos)
                    return (QValidator.Acceptable, value, pos)
            self._create_perspective_dialog.perspective_name_edit.setValidator(CustomValidator())

        # set default values
        self._create_perspective_dialog.perspective_name_edit.setText('')
        self._create_perspective_dialog.clone_checkbox.setChecked(True)
        self._create_perspective_dialog.clone_checkbox.setVisible(show_cloning)

        # show dialog and wait for it's return value
        return_value = self._create_perspective_dialog.exec_()
        if return_value == self._create_perspective_dialog.Rejected:
            return

        name = str(self._create_perspective_dialog.perspective_name_edit.text()).lstrip(
            self.HIDDEN_PREFIX)
        if name == '':
            QMessageBox.warning(
                self._menu_manager.menu,
                self.tr('Empty perspective name'),
                self.tr('The name of the perspective must be non-empty.'))
            return
        if name in self.perspectives:
            QMessageBox.warning(
                self._menu_manager.menu,
                self.tr('Duplicate perspective name'),
                self.tr('A perspective with the same name already exists.'))
            return
        return name

    def _create_perspective(self, name, clone_perspective=True):
        # convert from unicode
        name = str(name)
        if name.find('/') != -1:
            raise RuntimeError(
                'PerspectiveManager._create_perspective() name cannot contain forward slashes (/)')

        qDebug('PerspectiveManager._create_perspective(%s, %s)' % (name, clone_perspective))
        # add to list of perspectives
        self.perspectives.append(name)
        self._settings_proxy.set_value('', 'perspectives', self.perspectives)

        # save current settings
        if self._global_settings is not None and self._perspective_settings is not None:
            self._callback = self._create_perspective_continued
            self._callback_args = [name, clone_perspective]
            self.save_settings_signal.emit(self._global_settings, self._perspective_settings)
        else:
            self._create_perspective_continued(name, clone_perspective)

    def _create_perspective_continued(self, name, clone_perspective):
        # clone settings
        if clone_perspective:
            new_settings = self._get_perspective_settings(name)
            keys = self._perspective_settings.all_keys()
            for key in keys:
                value = self._perspective_settings.value(key)
                new_settings.set_value(key, value)

        # add and switch to perspective
        if not name.startswith(self.HIDDEN_PREFIX):
            self._add_perspective_action(name)

    def _add_perspective_action(self, name):
        if self._menu_manager is not None:
            # create action
            action = QAction(name, self._menu_manager.menu)
            action.setCheckable(True)
            self._perspective_mapper.setMapping(action, name)
            action.triggered.connect(self._perspective_mapper.map)

            # add action to menu
            self._menu_manager.add_item(action)
            # enable remove-action
            if self._menu_manager.count_items() > 1:
                self._remove_action.setEnabled(True)

    def _on_remove_perspective(self):
        # input dialog to choose perspective to be removed
        names = list(self.perspectives)
        names.remove(self._current_perspective)
        name, return_value = QInputDialog.getItem(
            self._menu_manager.menu, self._menu_manager.tr('Remove perspective'),
            self._menu_manager.tr('Select the perspective'), names, 0, False)
        # convert from unicode
        name = str(name)
        if return_value == QInputDialog.Rejected:
            return
        self._remove_perspective(name)

    def _remove_perspective(self, name):
        if name not in self.perspectives:
            raise UserWarning('unknown perspective: %s' % name)
        qDebug('PerspectiveManager._remove_perspective(%s)' % str(name))

        # remove from list of perspectives
        self.perspectives.remove(name)
        self._settings_proxy.set_value('', 'perspectives', self.perspectives)

        # remove settings
        settings = self._get_perspective_settings(name)
        settings.remove('')

        # remove from menu
        self._menu_manager.remove_item(name)

        # disable remove-action
        if self._menu_manager.count_items() < 2:
            self._remove_action.setEnabled(False)

    def _on_import_perspective(self):
        file_name, _ = QFileDialog.getOpenFileName(
            self._menu_manager.menu, self.tr('Import perspective from file'),
            self._file_path, self.tr('Perspectives (*.perspective)'))
        if file_name is None or file_name == '':
            return

        perspective_name = os.path.basename(file_name)
        suffix = '.perspective'
        if perspective_name.endswith(suffix):
            perspective_name = perspective_name[:-len(suffix)]
        if perspective_name in self.perspectives:
            perspective_name = self._choose_new_perspective_name(False)
            if perspective_name is None:
                return

        self.import_perspective_from_file(file_name, perspective_name)

    def import_perspective_from_file(self, path, perspective_name):
        self._file_path = os.path.dirname(path)
        # create clean perspective
        if perspective_name in self.perspectives:
            self._remove_perspective(perspective_name)
        self._create_perspective(perspective_name, clone_perspective=False)

        # read perspective from file
        file_handle = open(path, 'r')
        # data = eval(file_handle.read())
        data = json.loads(file_handle.read())
        self._convert_values(data, self._import_value)

        new_settings = self._get_perspective_settings(perspective_name)
        self._set_dict_on_settings(data, new_settings)

        self.switch_perspective(perspective_name, settings_changed=True, save_before=True)

    def _set_dict_on_settings(self, data, settings):
        """Set dictionary key-value pairs on Settings instance."""
        keys = data.get('keys', {})
        for key in keys:
            settings.set_value(key, keys[key])
        groups = data.get('groups', {})
        for group in groups:
            sub = settings.get_settings(group)
            self._set_dict_on_settings(groups[group], sub)

    def _on_export_perspective(self):
        save_file_name = os.path.join(
            self._file_path, self._current_perspective.lstrip(self.HIDDEN_PREFIX))
        suffix = '.perspective'
        if not save_file_name.endswith(suffix):
            save_file_name += suffix
        file_name, _ = QFileDialog.getSaveFileName(
            self._menu_manager.menu, self.tr('Export perspective to file'),
            save_file_name, self.tr('Perspectives (*.perspective)'))
        if file_name is None or file_name == '':
            return
        self._file_path = os.path.dirname(file_name)

        # trigger save of perspective before export
        self._callback = self._on_export_perspective_continued
        self._callback_args = [file_name]
        self.save_settings_signal.emit(self._global_settings, self._perspective_settings)

    def _on_export_perspective_continued(self, file_name):
        # convert every value
        data = self._get_dict_from_settings(self._perspective_settings)
        self._convert_values(data, self._export_value)

        # write perspective data to file
        file_handle = open(file_name, 'w')
        file_handle.write(json.dumps(data, indent=2, separators=(',', ': ')))
        file_handle.close()

    def _get_dict_from_settings(self, settings):
        """Convert data of Settings instance to dictionary."""
        keys = {}
        for key in settings.child_keys():
            keys[str(key)] = settings.value(key)
        groups = {}
        for group in settings.child_groups():
            sub = settings.get_settings(group)
            groups[str(group)] = self._get_dict_from_settings(sub)
        return {'keys': keys, 'groups': groups}

    def _convert_values(self, data, convert_function):
        keys = data.get('keys', {})
        for key in keys:
            keys[key] = convert_function(keys[key])
        groups = data.get('groups', {})
        for group in groups:
            self._convert_values(groups[group], convert_function)

    def _import_value(self, value):
        import QtCore  # noqa: F401
        if value['type'] == 'repr':
            return eval(value['repr'])
        elif value['type'] == 'repr(QByteArray.hex)':
            return QByteArray.fromHex(eval(value['repr(QByteArray.hex)']))
        raise RuntimeError(
            'PerspectiveManager._import_value() unknown serialization type (%s)' % value['type'])

    def _export_value(self, value):
        data = {}
        if value.__class__.__name__ == 'QByteArray':
            hex_value = value.toHex()
            data['repr(QByteArray.hex)'] = \
                self._strip_qt_binding_prefix(hex_value, repr(hex_value))
            data['type'] = 'repr(QByteArray.hex)'

            # add pretty print for better readability
            characters = ''
            for i in range(1, value.size(), 2):
                try:
                    character = value.at(i)
                    if isinstance(character, bytes):
                        character = character.decode('utf-8')
                    # output all non-control characters
                    if character >= ' ' and character <= '~':
                        characters += character
                    else:
                        characters += ' '
                except UnicodeDecodeError:
                    characters += ' '
            data['pretty-print'] = characters

        else:
            data['repr'] = self._strip_qt_binding_prefix(value, repr(value))
            data['type'] = 'repr'

        # verify that serialized data can be deserialized correctly
        reimported = self._import_value(data)
        if reimported != value:
            raise RuntimeError(
                'PerspectiveManager._export_value() stored value can not be restored (%s)' %
                type(value))

        return data

    def _strip_qt_binding_prefix(self, obj, data):
        """Strip binding specific prefix from type string."""
        parts = obj.__class__.__module__.split('.')
        if len(parts) > 1 and parts[1] == 'QtCore':
            prefix = '.'.join(parts[:2])
            data = data.replace(prefix, 'QtCore', 1)
        return data



# Copyright (c) 2011, Dirk Thomas, Dorian Scholz, TU Darmstadt
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
#   * Redistributions of source code must retain the above copyright
#     notice, this list of conditions and the following disclaimer.
#   * Redistributions in binary form must reproduce the above
#     copyright notice, this list of conditions and the following
#     disclaimer in the documentation and/or other materials provided
#     with the distribution.
#   * Neither the name of the TU Darmstadt nor the names of its
#     contributors may be used to endorse or promote products derived
#     from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

from qt_gui.dock_widget_container import DockWidgetContainer
from qt_gui.plugin_handler import PluginHandler
from qt_gui.window_changed_signaler import WindowChangedSignaler


class PluginHandlerContainer(PluginHandler):
    """Handler for creating a container."""

    def __init__(self, parent, main_window, instance_id, application_context, container_manager):
        super(PluginHandlerContainer, self).__init__(
            parent, main_window, instance_id, application_context, container_manager)
        self.setObjectName('PluginHandlerContainer')
        self._container = None

    def _load(self):
        self._container = DockWidgetContainer(
            self._container_manager, self._instance_id.serial_number)
        self._container.setObjectName(self._instance_id.tidy_str())
        title = self.tr('Container')
        if self._instance_id.serial_number > 1:
            title += ' (%d)' % self._instance_id.serial_number
        self._container.setWindowTitle(title)
        self._update_dock_widget_features(self._container)
        self._add_dock_widget_to_main_window(self._container)
        self._update_title_bar(self._container, True, True)
        signaler2 = WindowChangedSignaler(self._container, self._container)
        signaler2.hide_signal.connect(self._on_dock_widget_hide)
        signaler2.show_signal.connect(self._on_dock_widget_show)
        signaler2.window_title_changed_signal.connect(self._on_dock_widget_title_changed)
        # trigger to update initial window state
        signaler2.emit_all()
        self._widgets[self._container.main_window] = [self._container, None, signaler2]
        self._container_manager.add_container(self._container)
        self._emit_load_completed()

    def remove_widget(self, widget):
        self._container_manager.move_container_children_to_parent(self._container)
        self._container_manager.remove_container(self._container)
        super(PluginHandlerContainer, self).remove_widget(widget)
        self._container = None

    def _shutdown_plugin(self):
        self.emit_shutdown_plugin_completed()

    def _unload(self):
        self._emit_unload_completed()

    def _save_settings(self, plugin_settings, instance_settings):
        self.emit_save_settings_completed()

    def _restore_settings(self, plugin_settings, instance_settings):
        self.emit_restore_settings_completed()

    def _close_dock_widget(self, dock_widget):
        self._emit_close_plugin()



# Copyright (c) 2011, Dirk Thomas, Dorian Scholz, TU Darmstadt
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
#   * Redistributions of source code must retain the above copyright
#     notice, this list of conditions and the following disclaimer.
#   * Redistributions in binary form must reproduce the above
#     copyright notice, this list of conditions and the following
#     disclaimer in the documentation and/or other materials provided
#     with the distribution.
#   * Neither the name of the TU Darmstadt nor the names of its
#     contributors may be used to endorse or promote products derived
#     from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

from python_qt_binding.QtCore import QObject, QSignalMapper, Signal, Slot
from python_qt_binding.QtWidgets import QAction, QMenu

from qt_gui.icon_loader import get_icon
from qt_gui.menu_manager import MenuManager
from qt_gui.plugin_instance_id import PluginInstanceId


class PluginMenu(QObject):
    """Menu of available plugins to load and running plugin instances to unload."""

    load_plugin_signal = Signal(str)
    unload_plugin_signal = Signal(str)

    def __init__(self, menu_bar, plugin_manager):
        super(PluginMenu, self).__init__()
        self.setObjectName('PluginMenu')

        plugin_menu = menu_bar.addMenu(menu_bar.tr('&Plugins'))
        running_menu = menu_bar.addMenu(menu_bar.tr('&Running'))
        self._plugin_menu_manager = MenuManager(plugin_menu)
        self._plugin_mapper = QSignalMapper(plugin_menu)
        self._plugin_mapper.mapped[str].connect(self.load_plugin_signal)
        self._running_menu_manager = MenuManager(running_menu)
        action = QAction(
            ' Hidden action to work around QTBUG-52582', self._running_menu_manager.menu)
        action.setVisible(False)
        self._running_menu_manager.add_item(action)
        self._running_mapper = QSignalMapper(running_menu)
        self._running_mapper.mapped[str].connect(self.unload_plugin_signal)

        self._instances = {}

    def add_plugin(self, plugin_descriptor):
        base_path = plugin_descriptor.attributes().get('plugin_path')

        menu_manager = self._plugin_menu_manager
        # create submenus
        for group in plugin_descriptor.groups():
            label = group['label']
            if menu_manager.contains_menu(label):
                submenu = menu_manager.get_menu(label)
            else:
                submenu = QMenu(label, menu_manager.menu)
                menu_action = submenu.menuAction()
                self._enrich_action(menu_action, group, base_path)
                menu_manager.add_item(submenu)
            menu_manager = MenuManager(submenu)
        # create action
        action_attributes = plugin_descriptor.action_attributes()
        action = QAction(action_attributes['label'], menu_manager.menu)
        self._enrich_action(action, action_attributes, base_path)

        self._plugin_mapper.setMapping(action, plugin_descriptor.plugin_id())
        action.triggered.connect(self._plugin_mapper.map)

        not_available = plugin_descriptor.attributes().get('not_available')
        if not_available:
            action.setEnabled(False)
            action.setStatusTip(self.tr('Plugin is not available: %s') % not_available)

        # add action to menu
        menu_manager.add_item(action)

    def add_plugin_prefix(self, plugin_descriptor):
        action_attributes = plugin_descriptor.action_attributes()
        action = QAction(action_attributes['label'], self._plugin_menu_manager.menu)
        self._enrich_action(action, action_attributes)
        self._plugin_mapper.setMapping(action, plugin_descriptor.plugin_id())
        action.triggered.connect(self._plugin_mapper.map)
        self._plugin_menu_manager.add_prefix(action)

    def add_instance(self, plugin_descriptor, instance_id):
        action_attributes = plugin_descriptor.action_attributes()
        action = QAction(self._get_instance_label(
            str(instance_id)), self._running_menu_manager.menu)
        base_path = plugin_descriptor.attributes().get('plugin_path')
        self._enrich_action(action, action_attributes, base_path)

        self._running_mapper.setMapping(action, str(instance_id))
        action.triggered.connect(self._running_mapper.map)

        self._running_menu_manager.add_item(action)
        self._instances[instance_id] = action

    def remove_instance(self, instance_id):
        action = self._instances[instance_id]
        self._running_mapper.removeMappings(action)
        self._running_menu_manager.remove_item(action)

    @Slot(str, str)
    def update_plugin_instance_label(self, instance_id_str, label):
        instance_id = PluginInstanceId(instance_id=instance_id_str)
        action = self._instances[instance_id]
        action.setText(self._get_instance_label(label))

    def _get_instance_label(self, label):
        return self.tr('Close:') + ' ' + label

    def _enrich_action(self, action, action_attributes, base_path=None):
        if 'icon' in action_attributes and action_attributes['icon'] is not None:
            icon = get_icon(
                action_attributes['icon'], action_attributes.get('icontype', None), base_path)
            action.setIcon(icon)

        if 'statustip' in action_attributes:
            action.setStatusTip(action_attributes['statustip'])



# Copyright (c) 2011, Dirk Thomas, Dorian Scholz, TU Darmstadt
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
#   * Redistributions of source code must retain the above copyright
#     notice, this list of conditions and the following disclaimer.
#   * Redistributions in binary form must reproduce the above
#     copyright notice, this list of conditions and the following
#     disclaimer in the documentation and/or other materials provided
#     with the distribution.
#   * Neither the name of the TU Darmstadt nor the names of its
#     contributors may be used to endorse or promote products derived
#     from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

from qt_gui.plugin_handler_xembed_client import PluginHandlerXEmbedClient
from qt_gui.plugin_handler_xembed_container import PluginHandlerXEmbedContainer


class PluginHandlerXEmbed():
    """
    Handler for forwarding invocations between the framework and one `Plugin` instance.

    Handler uses peer-to-peer DBus connection. The both DBus endpoints are realized by the
    `PluginHandlerXEmbedContainer` and the `PluginHandlerXEmbedClient`.
    """

    def __init__(self, parent, main_window, instance_id,
                 application_context, container_manager, argv):
        dbus_object_path = '/PluginHandlerXEmbed/plugin/' + instance_id.tidy_str()
        if application_context.options.embed_plugin is None:
            self._handler = PluginHandlerXEmbedContainer(
                parent, main_window, instance_id, application_context,
                container_manager, argv, dbus_object_path)
        else:
            self._handler = PluginHandlerXEmbedClient(
                parent, main_window, instance_id, application_context,
                container_manager, argv, dbus_object_path)

    def __getattr__(self, name):
        return getattr(self._handler, name)



#!/usr/bin/env python

# Copyright (c) 2011, Dirk Thomas, Dorian Scholz, TU Darmstadt
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
#   * Redistributions of source code must retain the above copyright
#     notice, this list of conditions and the following disclaimer.
#   * Redistributions in binary form must reproduce the above
#     copyright notice, this list of conditions and the following
#     disclaimer in the documentation and/or other materials provided
#     with the distribution.
#   * Neither the name of the TU Darmstadt nor the names of its
#     contributors may be used to endorse or promote products derived
#     from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

from __future__ import print_function

from argparse import ArgumentParser, SUPPRESS
import os
import signal
import sys

from ament_index_python.resources import get_resource, has_resource


class Main(object):

    main_filename = None

    def __init__(self, qtgui_path, invoked_filename=None, settings_filename=None):
        self._qtgui_path = qtgui_path
        if invoked_filename is None:
            invoked_filename = os.path.abspath(__file__)
        Main.main_filename = invoked_filename
        if settings_filename is None:
            settings_filename = 'qt_gui'
        self._settings_filename = settings_filename

        self.plugin_providers = []
        self._options = None

        # check if DBus is available
        self._dbus_available = False
        try:
            # use qt/glib mainloop integration to get dbus mainloop working
            from dbus.mainloop.glib import DBusGMainLoop
            DBusGMainLoop(set_as_default=True)
            import dbus
            try:
                # before being able to check if a session bus is available the dbus
                # mainloop must be set up
                dbus.SessionBus()
                self._dbus_available = True
            except dbus.exceptions.DBusException:
                pass
        except ImportError:
            pass

    def add_arguments(self, parser, standalone=False, plugin_argument_provider=None):
        common_group = parser.add_argument_group('Options for GUI instance')
        common_group.add_argument(
            '-b', '--qt-binding', dest='qt_binding', type=str, metavar='BINDING',
            help='choose Qt bindings to be used [pyqt|pyside]')
        common_group.add_argument(
            '--clear-config', dest='clear_config', default=False, action='store_true',
            help='clear the configuration (including all perspectives and plugin settings)')
        if not standalone:
            common_group.add_argument(
                '-f', '--freeze-layout', dest='freeze_layout', action='store_true',
                help='freeze the layout of the GUI (prevent rearranging widgets, disable '
                     'undock/redock)')
        common_group.add_argument(
            '--force-discover', dest='force_discover', default=False, action='store_true',
            help='force a rediscover of plugins')
        common_group.add_argument(
            '-h', '--help', action='help', help='show this help message and exit')
        common_group.add_argument(
                '-ht', '--hide-title', dest='hide_title', action='store_true',
                help='hide the title label, the icon, and the help button (combine with -l and -f '
                     'to eliminate the entire title bar and reclaim the space)')
        if not standalone:
            common_group.add_argument(
                '-l', '--lock-perspective', dest='lock_perspective', action='store_true',
                help='lock the GUI to the used perspective (hide menu bar and close buttons of '
                     'plugins)')

            common_group.add_argument(
                '-p', '--perspective', dest='perspective', type=str, metavar='PERSPECTIVE',
                help='start with this named perspective')
            common_group.add_argument(
                '--perspective-file', dest='perspective_file', type=str,
                metavar='PERSPECTIVE_FILE', help='start with a perspective loaded from a file')
        common_group.add_argument(
            '--reload-import', dest='reload_import', default=False, action='store_true',
            help='reload every imported module')
        if not standalone:
            common_group.add_argument(
                '-s', '--standalone', dest='standalone_plugin', type=str, metavar='PLUGIN',
                help='start only this plugin (implies -l). To pass arguments to the plugin use '
                     '--args')
        common_group.add_argument(
            '-t', '--on-top', dest='on_top', default=False, action='store_true',
            help='set window mode to always on top')
        common_group.add_argument(
            '-v', '--verbose', dest='verbose', default=False, action='store_true',
            help='output qDebug messages')

        if not standalone:
            common_group.add_argument(
                '--args', dest='plugin_args', nargs='*', type=str,
                help='arbitrary arguments which are passes to the plugin '
                     '(only with -s, --command-start-plugin or --embed-plugin). '
                     'It must be the last option since it collects all following options.')

            group = parser.add_argument_group(
                'Options to query information without starting a GUI instance',
                'These options can be used to query information about valid arguments for various '
                'options.')
            group.add_argument(
                '--list-perspectives', dest='list_perspectives', action='store_true',
                help='list available perspectives')
            group.add_argument(
                '--list-plugins', dest='list_plugins', action='store_true',
                help='list available plugins')
            parser.add_argument_group(group)

            group = parser.add_argument_group(
                'Options to operate on a running GUI instance',
                'These options can be used to perform actions on a running GUI instance.')
            group.add_argument(
                '--command-pid', dest='command_pid', type=int, metavar='PID',
                help='pid of the GUI instance to operate on, defaults to oldest running GUI '
                'instance')
            group.add_argument(
                '--command-start-plugin', dest='command_start_plugin', type=str, metavar='PLUGIN',
                help='start plugin')
            group.add_argument(
                '--command-switch-perspective', dest='command_switch_perspective', type=str,
                metavar='PERSPECTIVE', help='switch perspective')
            if not self._dbus_available:
                group.description = 'These options are not available since DBus is not available!'
                for o in group._group_actions:
                    o.help = SUPPRESS
            parser.add_argument_group(group)

            group = parser.add_argument_group(
                'Special options for embedding widgets from separate processes',
                'These options should never be used on the CLI but only from the GUI code itself.')
            group.add_argument(
                '--embed-plugin', dest='embed_plugin', type=str, metavar='PLUGIN',
                help='embed a plugin into an already running GUI instance (requires all other '
                     '--embed-* options)')
            group.add_argument(
                '--embed-plugin-serial', dest='embed_plugin_serial', type=int, metavar='SERIAL',
                help='serial number of plugin to be embedded '
                     '(requires all other --embed-* options)')
            group.add_argument(
                '--embed-plugin-address', dest='embed_plugin_address', type=str, metavar='ADDRESS',
                help='dbus server address of the GUI instance to embed plugin into '
                     '(requires all other --embed-* options)')
            for o in group._group_actions:
                o.help = SUPPRESS
            parser.add_argument_group(group)

        if plugin_argument_provider:
            plugin_argument_provider(parser)

        return common_group

    def _add_plugin_providers(self):
        pass

    def _add_reload_paths(self, reload_importer):
        reload_importer.add_reload_path(os.path.join(os.path.dirname(__file__), *('..',) * 4))

    def _set_theme_if_necessary(self):
        from python_qt_binding.QtGui import QIcon
        # Always prefer tango_icons_vendor
        package_path = has_resource('packages', 'tango_icons_vendor')
        if package_path:
            icon_paths = QIcon.themeSearchPaths()
            icon_paths = [
                os.path.join(package_path, 'share', 'tango_icons_vendor', 'resource', 'icons')
            ] + icon_paths
            QIcon.setThemeSearchPaths(icon_paths)

        # Use Tango if possible, fall back to system default
        original_theme = QIcon.themeName()
        if original_theme == 'Tango':
            # We don't have any back-up behaviour in case Tango is the default
            return
        # Set Tango as default and try to get an icon
        QIcon.setThemeName('Tango')
        if QIcon.fromTheme('document-save').isNull():
            # When Tango icon can't be found, use the system default theme
            QIcon.setThemeName(original_theme)
            return
        # In case we overruled the default theme to be Tango, set fallback theme to the original
        # theme. Any original fallback theme is dropped in this case.
        QIcon.setFallbackThemeName(original_theme)

    def create_application(self, argv):
        from python_qt_binding.QtCore import Qt
        from python_qt_binding.QtWidgets import QApplication
        app = QApplication(argv)
        app.setAttribute(Qt.AA_DontShowIconsInMenus, False)
        return app

    def main(self, argv=None, standalone=None, plugin_argument_provider=None,
             plugin_manager_settings_prefix='', help_text=None):
        if argv is None:
            argv = sys.argv

        # extract --args and everything behind manually since argparse can not handle that
        arguments = argv[1:]

        # extract plugin specific args when not being invoked in standalone mode programmatically
        if not standalone:
            plugin_args = []
            if '--args' in arguments:
                index = arguments.index('--args')
                plugin_args = arguments[index + 1:]
                arguments = arguments[0:index + 1]

        parser = ArgumentParser(os.path.basename(Main.main_filename), add_help=False)
        self.add_arguments(parser, standalone=bool(standalone),
                           plugin_argument_provider=plugin_argument_provider)
        self._options = parser.parse_args(arguments)

        if standalone:
            # rerun parsing to separate common arguments from plugin specific arguments
            parser = ArgumentParser(os.path.basename(Main.main_filename), add_help=False)
            self.add_arguments(parser, standalone=bool(standalone))
            self._options, plugin_args = parser.parse_known_args(arguments)
        self._options.multi_process = False  # not supported anymore
        self._options.plugin_args = plugin_args

        # set default values for options not available in standalone mode
        if standalone:
            self._options.freeze_layout = False
            self._options.lock_perspective = False
            self._options.perspective = None
            self._options.perspective_file = None
            self._options.standalone_plugin = standalone
            self._options.list_perspectives = False
            self._options.list_plugins = False
            self._options.command_pid = None
            self._options.command_start_plugin = None
            self._options.command_switch_perspective = None
            self._options.embed_plugin = None
            self._options.embed_plugin_serial = None
            self._options.embed_plugin_address = None

        # check option dependencies
        try:
            if self._options.plugin_args and \
                    not self._options.standalone_plugin and \
                    not self._options.command_start_plugin and \
                    not self._options.embed_plugin:
                raise RuntimeError(
                    'Option --args can only be used together with either --standalone, '
                    '--command-start-plugin or --embed-plugin option')

            if self._options.freeze_layout and not self._options.lock_perspective:
                raise RuntimeError(
                    'Option --freeze_layout can only be used together with the '
                    '--lock_perspective option')

            list_options = (self._options.list_perspectives, self._options.list_plugins)
            list_options_set = [opt for opt in list_options if opt is not False]
            if len(list_options_set) > 1:
                raise RuntimeError('Only one --list-* option can be used at a time')

            command_options = (
                self._options.command_start_plugin, self._options.command_switch_perspective)
            command_options_set = [opt for opt in command_options if opt is not None]
            if len(command_options_set) > 0 and not self._dbus_available:
                raise RuntimeError(
                    'Without DBus support the --command-* options are not available')
            if len(command_options_set) > 1:
                raise RuntimeError(
                    'Only one --command-* option can be used at a time (except --command-pid '
                    'which is optional)')
            if len(command_options_set) == 0 and self._options.command_pid is not None:
                raise RuntimeError(
                    'Option --command_pid can only be used together with an other --command-* '
                    'option')

            embed_options = (
                self._options.embed_plugin, self._options.embed_plugin_serial,
                self._options.embed_plugin_address)
            embed_options_set = [opt for opt in embed_options if opt is not None]
            if len(command_options_set) > 0 and not self._dbus_available:
                raise RuntimeError('Without DBus support the --embed-* options are not available')
            if len(embed_options_set) > 0 and len(embed_options_set) < len(embed_options):
                raise RuntimeError("Missing option(s) - all '--embed-*' options must be set")

            if len(embed_options_set) > 0 and self._options.clear_config:
                raise RuntimeError(
                    'Option --clear-config can only be used without any --embed-* option')

            groups = (list_options_set, command_options_set, embed_options_set)
            groups_set = [opt for opt in groups if len(opt) > 0]
            if len(groups_set) > 1:
                raise RuntimeError(
                    'Options from different groups (--list, --command, --embed) can not be used '
                    'together')

            perspective_options = (self._options.perspective, self._options.perspective_file)
            perspective_options_set = [opt for opt in perspective_options if opt is not None]
            if len(perspective_options_set) > 1:
                raise RuntimeError('Only one --perspective-* option can be used at a time')

            if self._options.perspective_file is not None and \
                    not os.path.isfile(self._options.perspective_file):
                raise RuntimeError('Option --perspective-file must reference existing file')

        except RuntimeError as e:
            print(str(e))
            # parser.parse_args(['--help'])
            # calling --help will exit
            return 1

        # set implicit option dependencies
        if self._options.standalone_plugin is not None:
            self._options.lock_perspective = True

        # create application context containing various relevant information
        from qt_gui.application_context import ApplicationContext
        context = ApplicationContext()
        context.qtgui_path = self._qtgui_path
        context.options = self._options

        if self._dbus_available:
            from dbus import DBusException, Interface, SessionBus

        # non-special applications provide various dbus interfaces
        if self._dbus_available:
            context.provide_app_dbus_interfaces = len(groups_set) == 0
            context.dbus_base_bus_name = 'org.ros.qt_gui'
            if context.provide_app_dbus_interfaces:
                context.dbus_unique_bus_name = context.dbus_base_bus_name + '.pid%d' % os.getpid()

                # provide pid of application via dbus
                from qt_gui.application_dbus_interface import ApplicationDBusInterface
                _dbus_server = ApplicationDBusInterface(context.dbus_base_bus_name)  # noqa: F841

        # determine host bus name, either based on pid given on command line or
        # via dbus application interface if any other instance is available
        if len(command_options_set) > 0 or len(embed_options_set) > 0:
            host_pid = None
            if self._options.command_pid is not None:
                host_pid = self._options.command_pid
            else:
                try:
                    remote_object = SessionBus().get_object(
                        context.dbus_base_bus_name, '/Application')
                except DBusException:
                    pass
                else:
                    remote_interface = Interface(
                        remote_object, context.dbus_base_bus_name + '.Application')
                    host_pid = remote_interface.get_pid()
            if host_pid is not None:
                context.dbus_host_bus_name = context.dbus_base_bus_name + '.pid%d' % host_pid

        # execute command on host application instance
        if len(command_options_set) > 0:
            if self._options.command_start_plugin is not None:
                try:
                    remote_object = SessionBus().get_object(
                        context.dbus_host_bus_name, '/PluginManager')
                except DBusException:
                    (rc, msg) = (1, 'unable to communicate with GUI instance "%s"' %
                                 context.dbus_host_bus_name)
                else:
                    remote_interface = Interface(
                        remote_object, context.dbus_base_bus_name + '.PluginManager')
                    (rc, msg) = remote_interface.start_plugin(
                        self._options.command_start_plugin, ' '.join(self._options.plugin_args))
                if rc == 0:
                    print('qt_gui_main() started plugin "%s" in GUI "%s"' %
                          (msg, context.dbus_host_bus_name))
                else:
                    print('qt_gui_main() could not start plugin "%s" in GUI "%s": %s' %
                          (self._options.command_start_plugin, context.dbus_host_bus_name, msg))
                return rc
            elif self._options.command_switch_perspective is not None:
                remote_object = SessionBus().get_object(
                    context.dbus_host_bus_name, '/PerspectiveManager')
                remote_interface = Interface(
                    remote_object, context.dbus_base_bus_name + '.PerspectiveManager')
                remote_interface.switch_perspective(self._options.command_switch_perspective)
                print('qt_gui_main() switched to perspective "%s" in GUI "%s"' %
                      (self._options.command_switch_perspective, context.dbus_host_bus_name))
                return 0
            raise RuntimeError('Unknown command not handled')

        # choose selected or default qt binding
        setattr(sys, 'SELECT_QT_BINDING', self._options.qt_binding)
        from python_qt_binding import QT_BINDING

        from python_qt_binding.QtCore import (qDebug, qInstallMessageHandler,
                                              QSettings, Qt, QtCriticalMsg, QtDebugMsg)
        from python_qt_binding.QtCore import QtFatalMsg, QTimer, QtWarningMsg

        from python_qt_binding.QtGui import QIcon
        from python_qt_binding.QtWidgets import QAction

        from qt_gui.about_handler import AboutHandler
        from qt_gui.composite_plugin_provider import CompositePluginProvider
        from qt_gui.container_manager import ContainerManager
        from qt_gui.help_provider import HelpProvider
        from qt_gui.icon_loader import get_icon
        from qt_gui.main_window import MainWindow
        from qt_gui.minimized_dock_widgets_toolbar import MinimizedDockWidgetsToolbar
        from qt_gui.perspective_manager import PerspectiveManager
        from qt_gui.plugin_manager import PluginManager

        # TODO PySide2 segfaults when invoking this custom message handler atm
        if QT_BINDING != 'pyside':
            def message_handler(type_, context, msg):
                colored_output = 'TERM' in os.environ and 'ANSI_COLORS_DISABLED' not in os.environ
                cyan_color = '\033[36m' if colored_output else ''
                red_color = '\033[31m' if colored_output else ''
                reset_color = '\033[0m' if colored_output else ''
                if type_ == QtDebugMsg and self._options.verbose:
                    print(msg, file=sys.stderr)
                elif type_ == QtWarningMsg:
                    print(cyan_color + msg + reset_color, file=sys.stderr)
                elif type_ == QtCriticalMsg:
                    print(red_color + msg + reset_color, file=sys.stderr)
                elif type_ == QtFatalMsg:
                    print(red_color + msg + reset_color, file=sys.stderr)
                    sys.exit(1)
            qInstallMessageHandler(message_handler)

        app = self.create_application(argv)

        self._set_theme_if_necessary()

        settings = QSettings(
            QSettings.IniFormat, QSettings.UserScope, 'ros.org', self._settings_filename)
        if len(embed_options_set) == 0:
            if self._options.clear_config:
                settings.clear()

            main_window = MainWindow(help_text)
            if self._options.on_top:
                main_window.setWindowFlags(Qt.WindowStaysOnTopHint)

            main_window.statusBar()

            def sigint_handler(*args):
                qDebug('\nsigint_handler()')
                main_window.close()
            signal.signal(signal.SIGINT, sigint_handler)
            # the timer enables triggering the sigint_handler
            timer = QTimer()
            timer.start(500)
            timer.timeout.connect(lambda: None)

            if not self._options.lock_perspective:
                menu_bar = main_window.menuBar()
                file_menu = menu_bar.addMenu(menu_bar.tr('&File'))
                action = QAction(file_menu.tr('&Quit'), file_menu)
                action.setIcon(QIcon.fromTheme('application-exit'))
                action.triggered.connect(main_window.close)
                file_menu.addAction(action)
            else:
                menu_bar = None

        else:
            app.setQuitOnLastWindowClosed(False)

            main_window = None
            menu_bar = None

        self._add_plugin_providers()

        # setup plugin manager
        plugin_provider = CompositePluginProvider(self.plugin_providers)
        plugin_manager = PluginManager(
            plugin_provider, settings, context, settings_prefix=plugin_manager_settings_prefix)

        if self._options.list_plugins:
            # output available plugins
            print('\n'.join(sorted(plugin_manager.get_plugins().values())))
            return 0

        help_provider = HelpProvider()
        plugin_manager.plugin_help_signal.connect(help_provider.plugin_help_request)

        # setup perspective manager
        if main_window is not None:
            perspective_manager = PerspectiveManager(settings, context)

            if self._options.list_perspectives:
                # output available perspectives
                print('\n'.join(sorted(perspective_manager.perspectives)))
                return 0
        else:
            perspective_manager = None

        if main_window is not None:
            container_manager = ContainerManager(main_window, plugin_manager)
            plugin_manager.set_main_window(main_window, menu_bar, container_manager)
            plugin_manager.plugin_load_unload_signal.connect(main_window.showHelpWidget)

            if not self._options.freeze_layout:
                minimized_dock_widgets_toolbar = MinimizedDockWidgetsToolbar(
                    container_manager, main_window)
                main_window.addToolBar(Qt.BottomToolBarArea, minimized_dock_widgets_toolbar)
                plugin_manager.set_minimized_dock_widgets_toolbar(minimized_dock_widgets_toolbar)

        if menu_bar is not None:
            perspective_menu = menu_bar.addMenu(menu_bar.tr('P&erspectives'))
            perspective_manager.set_menu(perspective_menu)

        # connect various signals and slots
        if perspective_manager is not None and main_window is not None:
            # signal changed perspective to update window title
            perspective_manager.perspective_changed_signal.connect(main_window.perspective_changed)
            # signal new settings due to changed perspective
            perspective_manager.save_settings_signal.connect(main_window.save_settings)
            perspective_manager.restore_settings_signal.connect(main_window.restore_settings)
            perspective_manager.restore_settings_without_plugin_changes_signal.connect(
                main_window.restore_settings)

        if perspective_manager is not None and plugin_manager is not None:
            perspective_manager.save_settings_signal.connect(plugin_manager.save_settings)
            plugin_manager.save_settings_completed_signal.connect(
                perspective_manager.save_settings_completed)
            perspective_manager.restore_settings_signal.connect(plugin_manager.restore_settings)
            perspective_manager.restore_settings_without_plugin_changes_signal.connect(
                plugin_manager.restore_settings_without_plugins)

        if plugin_manager is not None and main_window is not None:
            # signal before changing plugins to save window state
            plugin_manager.plugins_about_to_change_signal.connect(main_window.save_setup)
            # signal changed plugins to restore window state
            plugin_manager.plugins_changed_signal.connect(main_window.restore_state)
            # signal save settings to store plugin setup on close
            main_window.save_settings_before_close_signal.connect(plugin_manager.close_application)
            # signal save and shutdown called for all plugins, trigger closing main window again
            plugin_manager.close_application_signal.connect(
                main_window.close, type=Qt.QueuedConnection)

        if main_window is not None and menu_bar is not None:
            about_handler = AboutHandler(context.qtgui_path, main_window)
            help_menu = menu_bar.addMenu(menu_bar.tr('&Help'))
            action = QAction(help_menu.tr('&About'), help_menu)
            action.setIcon(QIcon.fromTheme('help-about'))
            action.triggered.connect(about_handler.show)
            help_menu.addAction(action)

        # set initial size - only used without saved configuration
        if main_window is not None:
            # Adjust size to fit the widget if standalone (e.g. single plugin)
            if standalone:
                main_window.adjustSize()
            # On "clean" startup set some size to fully display the menu bar
            else:
                main_window.resize(600, 450)
                main_window.move(100, 100)

        # ensure that qt_gui/src is in sys.path
        src_path = os.path.realpath(os.path.join(os.path.dirname(__file__), '..'))
        if src_path not in sys.path:
            sys.path.append(src_path)

        # load specific plugin
        plugin = None
        plugin_serial = None
        if self._options.embed_plugin is not None:
            plugin = self._options.embed_plugin
            plugin_serial = self._options.embed_plugin_serial
        elif self._options.standalone_plugin is not None:
            plugin = self._options.standalone_plugin
            plugin_serial = 0
        if plugin is not None:
            plugins = plugin_manager.find_plugins_by_name(plugin)
            if len(plugins) == 0:
                print('qt_gui_main() found no plugin matching "%s"' % plugin)
                print('try passing the option "--force-discover"')
                return 1
            elif len(plugins) > 1:
                print('qt_gui_main() found multiple plugins matching "%s"\n%s' %
                      (plugin, '\n'.join(plugins.values())))
                return 1
            plugin = list(plugins.keys())[0]

        qDebug('QtBindingHelper using %s' % QT_BINDING)

        plugin_manager.discover()

        if self._options.reload_import:
            qDebug('ReloadImporter() automatically reload all subsequent imports')
            from qt_gui.reload_importer import ReloadImporter
            _reload_importer = ReloadImporter()
            self._add_reload_paths(_reload_importer)
            _reload_importer.enable()

        # switch perspective
        if perspective_manager is not None:
            if plugin:
                perspective_manager.set_perspective(plugin, hide_and_without_plugin_changes=True)
            elif self._options.perspective_file:
                perspective_manager.import_perspective_from_file(
                    self._options.perspective_file,
                    perspective_manager.HIDDEN_PREFIX + os.path.basename(
                        self._options.perspective_file))
            else:
                perspective_manager.set_perspective(self._options.perspective)

        # load specific plugin
        if plugin:
            plugin_manager.load_plugin(plugin, plugin_serial, self._options.plugin_args)
            running = plugin_manager.is_plugin_running(plugin, plugin_serial)
            if not running:
                return 1
            if self._options.standalone_plugin:
                # use icon of standalone plugin (if available) for application
                plugin_descriptor = plugin_manager.get_plugin_descriptor(plugin)
                action_attributes = plugin_descriptor.action_attributes()
                if 'icon' in action_attributes and action_attributes['icon'] is not None:
                    base_path = plugin_descriptor.attributes().get('plugin_path')
                    try:
                        icon = get_icon(
                            action_attributes['icon'],
                            action_attributes.get('icontype', None),
                            base_path)
                    except UserWarning:
                        pass
                    else:
                        app.setWindowIcon(icon)

        if main_window is not None:
            main_window.show()
            if sys.platform == 'darwin':
                main_window.raise_()

        return app.exec_()


if __name__ == '__main__':
    _, qtgui_path = get_resource('packages', 'qt_gui')
    main = Main(qtgui_path)
    sys.exit(main.main())
